"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Contains unit and integration tests for the related functionality.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

import asyncio

from notolog.async_highlighter import AsyncHighlighter

import pytest
import pytest_asyncio


class TestQtAsync:

    @pytest.fixture
    def async_highlighter_obj(self):
        # To avoid: There is no current event loop when self.loop = asyncio.get_event_loop()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        # Fixture to create and return object instance
        async_highlighter = AsyncHighlighter(callback=lambda check: check)
        yield async_highlighter

    @pytest_asyncio.fixture
    async def cleanup_tasks(self):
        yield
        # Cancel and clean up all pending asyncio tasks
        for task in asyncio.all_tasks():
            if task is not asyncio.current_task():
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

    @pytest.mark.asyncio
    async def test_ui_interaction(self, mocker, async_highlighter_obj, cleanup_tasks):
        # Get callback method to check
        mock_rehighlight = mocker.patch.object(async_highlighter_obj, 'rehighlight_task_callback')

        # Check local task pool is empty
        assert len(async_highlighter_obj.rehighlight_tasks) == 0

        # Queue re-highlight tasks
        task0 = await async_highlighter_obj.rehighlight_in_queue()
        mock_rehighlight.assert_called_once_with(task0)
        # One more
        task1 = await async_highlighter_obj.rehighlight_in_queue()
        mock_rehighlight.assert_called_with(task1)
        # Another one more
        task2 = await async_highlighter_obj.rehighlight_in_queue()
        mock_rehighlight.assert_called_with(task2)
        # Add more tasks
        await async_highlighter_obj.rehighlight_in_queue()
        await async_highlighter_obj.rehighlight_in_queue()
        await async_highlighter_obj.rehighlight_in_queue()
        await async_highlighter_obj.rehighlight_in_queue()
        await async_highlighter_obj.rehighlight_in_queue()
        # Max count in a short period of time is 3
        assert mock_rehighlight.call_count == 3

        # Checking of local task pool doesn't work properly if run from IDE, but works from CLI
        assert len(async_highlighter_obj.rehighlight_tasks) == 3

        # The cleanup_tasks fixture is used implicitly; no need to assert its presence
        # assert cleanup_tasks



# German lexemes toolbar.py
lexemes = {
    "actions_label_new_document": "Neues Dokument",
    "actions_accessible_name_new_document": "Neues Dokument",
    "actions_label_open": "Öffnen",
    "actions_accessible_name_open": "Dokument öffnen",
    "actions_label_save": "Speichern",
    "actions_accessible_name_save": "Dokument speichern",
    "actions_label_save_as": "Speichern unter",
    "actions_accessible_name_save_as": "Kopie des Dokuments speichern",
    "actions_label_sync": "Synchronisieren",
    "actions_accessible_name_sync": "Dokument mit externem Speicher synchronisieren",
    "actions_label_edit": "Bearbeiten",
    "actions_accessible_name_edit": "Datei bearbeiten",
    "actions_label_source": "Quelle",
    "actions_accessible_name_source": "Quellcode",
    "actions_label_encrypt": "Verschlüsseln",
    "actions_accessible_name_encrypt": "Datei verschlüsseln",
    "actions_label_decrypt": "Entschlüsseln",
    "actions_accessible_name_decrypt": "Datei entschlüsseln",
    "actions_label_bold": "Fett",
    "actions_accessible_name_bold": "Text fett formatieren",
    "actions_label_italic": "Kursiv",
    "actions_accessible_name_italic": "Text kursiv formatieren",
    "actions_label_underline": "Unterstreichen",
    "actions_accessible_name_underline": "Text unterstreichen",
    "actions_label_strikethrough": "Durchstreichen",
    "actions_accessible_name_strikethrough": "Text durchstreichen",
    "actions_label_blockquote": "Blockzitat",
    "actions_accessible_name_blockquote": "Text als Blockzitat formatieren",
    "actions_label_color_picker": "Farbwähler",
    "actions_accessible_name_color_picker": "Textfarbe wählen",
    "actions_label_ai_assistant": "KI-Assistent",
    "actions_accessible_name_ai_assistant": "KI-Assistent ausführen",
    "actions_label_about": "Über",
    "actions_accessible_name_about": "Über die App",
    "actions_label_exit": "Beenden",
    "actions_accessible_name_exit": "App beenden",
    "actions_label_settings": "Einstellungen",
    "actions_accessible_name_settings": "App-Einstellungen",

    "search_input_label": "Suchen",
    "search_input_placeholder_text": "Text im Dokument suchen",
    "search_input_accessible_description": "Sucheingabefeld",
    "search_buttons_label_clear": "Suche löschen",
    "search_buttons_accessible_name_clear": "Sucheingabefeld löschen",
    "search_buttons_label_prev": "Vorherige Suche",
    "search_buttons_accessible_name_prev": "Vorheriges Suchergebnis",
    "search_buttons_label_next": "Nächste Suche",
    "search_buttons_accessible_name_next": "Nächstes Suchergebnis",
    "search_case_sensitive_accessible_name": "Groß-/Kleinschreibung beachten",
    "search_case_sensitive_tooltip": "Suche unter Berücksichtigung der Groß-/Kleinschreibung",
}



# Korean lexemes settings_dialog.py
lexemes = {
    "tab_openai_api_config": "OpenAI API",

    "module_openai_api_label": "OpenAI API",
    "module_openai_api_url_label": "API URL",
    "module_openai_api_url_input_placeholder_text": "API URL",
    "module_openai_api_url_input_accessible_description":
        "OpenAI API의 URL은 API 엔드포인트의 주소이며, 서비스와 버전에 따라 달라질 수 있습니다.\n"
        "AI 어시스턴트는 대화형 채팅 기능 또는 텍스트 완성을 위해 해당 엔드포인트를 사용합니다.\n"
        "현재 URL을 확인하려면 OpenAI API의 공식 문서를 참조하세요.",
    "module_openai_api_key_label": "API 키",
    "module_openai_api_key_input_placeholder_text": "API 키",
    "module_openai_api_key_input_accessible_description":
        "OpenAI API 키는 API 엔드포인트에 대한 요청을 인증하는 데 사용되는 비밀 토큰입니다.",
    "module_openai_api_supported_models_label": "지원 모델",
    "module_openai_api_model_names_combo_placeholder_text": "모델 선택",
    "module_openai_api_model_names_combo_accessible_description":
        "채팅 대화에 지원되는 모델 중에서 선택합니다.",

    "module_openai_api_base_system_prompt_label": "시스템 프롬프트",
    "module_openai_api_base_system_prompt_edit_placeholder_text": "각 요청에 앞서는 기본 시스템 프롬프트",
    "module_openai_api_base_system_prompt_edit_accessible_description":
        "각 요청에 앞서는 기본 시스템 프롬프트입니다.\n"
        "보통 지시사항이나 역할 특성을 포함한 일반 텍스트입니다.",

    "module_openai_api_base_response_temperature_label": "온도: {temperature}",
    "module_openai_api_base_response_temperature_input_accessible_description":
        "모델 응답의 무작위성을 조정합니다. 높은 값은 더 다양한 출력을 생성하고,\n"
        "낮은 값은 응답을 더 예측 가능하게 만듭니다.",

    "module_openai_api_base_response_max_tokens_label": "최대 응답 토큰 수",
    "module_openai_api_base_response_max_tokens_input_accessible_description":
        "응답에서 받을 수 있는 최대 토큰 수를 설정합니다. 이에는 단어와 구두점이 포함되며,\n"
        "출력의 길이를 제어합니다.",

    "module_openai_api_config_prompt_history_size_label": "프롬프트 이력 크기",
    "module_openai_api_config_prompt_history_size_input_accessible_description":
        "시스템이 참조용으로 유지하는 프롬프트 이력의 항목 수를 제어합니다.\n"
        "0의 값은 무제한 항목을 허용합니다."
}



# Spanish lexemes settings_dialog.py
lexemes = {
    "tab_openai_api_config": "API de OpenAI",

    "module_openai_api_label": "API de OpenAI",
    "module_openai_api_url_label": "URL de la API",
    "module_openai_api_url_input_placeholder_text": "URL de la API",
    "module_openai_api_url_input_accessible_description":
        "La URL de la API de OpenAI es la dirección del punto de conexión de la API, que puede variar según el servicio\n"
        "y la versión. El Asistente de IA utiliza la correspondiente para la funcionalidad de chat conversacional\n"
        "o completaciones de texto. Consulte la documentación oficial de la API de OpenAI para obtener la URL actual.",
    "module_openai_api_key_label": "Clave API",
    "module_openai_api_key_input_placeholder_text": "Clave API",
    "module_openai_api_key_input_accessible_description":
        "La clave API de OpenAI es un token secreto que se utiliza para autenticar solicitudes\n"
        "al punto de conexión de la API.",
    "module_openai_api_supported_models_label": "Modelos compatibles",
    "module_openai_api_model_names_combo_placeholder_text": "Elegir un modelo",
    "module_openai_api_model_names_combo_accessible_description":
        "Seleccione de los modelos compatibles para conversaciones de chat.",

    "module_openai_api_base_system_prompt_label": "Prompt del sistema",
    "module_openai_api_base_system_prompt_edit_placeholder_text": "Prompt del sistema base que precede a cada solicitud",
    "module_openai_api_base_system_prompt_edit_accessible_description":
        "Un prompt del sistema base que precede a cada solicitud.\n"
        "Por lo general, es texto plano con instrucciones o características de rol.",

    "module_openai_api_base_response_temperature_label": "Temperatura: {temperature}",
    "module_openai_api_base_response_temperature_input_accessible_description":
        "Ajusta la aleatoriedad de las respuestas del modelo. Valores más altos producen salidas más variadas, "
        "mientras que valores más bajos hacen que las respuestas sean más predecibles.",

    "module_openai_api_base_response_max_tokens_label": "Máximos tokens de respuesta",
    "module_openai_api_base_response_max_tokens_input_accessible_description":
        "Número máximo de tokens que se recibirán en una respuesta, como palabras y puntuación, "
        "controlando la longitud de la salida.",

    "module_openai_api_config_prompt_history_size_label": "Tamaño del historial de prompts",
    "module_openai_api_config_prompt_history_size_input_accessible_description":
        "Controla el número de entradas en el historial de prompts que el sistema retiene para referencia.\n"
        "Un valor cero permite entradas ilimitadas."
}



# French lexemes common.py
lexemes = {
    "module_llama_cpp_name": "Module llama.cpp",

    "module_llama_cpp_model_exception": "Erreur modèle non trouvé : {error_msg}",
    "module_llama_cpp_task_exception": "Impossible d'initialiser le générateur de modèle : {error_msg}",
}



# French lexemes settings_dialog.py
lexemes = {
    "tab_ondevice_llm_config": "LLM sur Appareil",

    "module_ondevice_llm_config_label": "Modèle LLM sur Appareil",
    "module_ondevice_llm_config_path_label": "Emplacement du modèle ONNX",
    "module_ondevice_llm_config_path_input_placeholder_text": "Chemin du répertoire du modèle",
    "module_ondevice_llm_config_path_input_accessible_description":
        "Un champ de saisie avec un sélecteur pour spécifier le chemin vers le répertoire du modèle où se trouvent\n"
        "les fichiers ONNX. Les modèles pris en charge sont au format ONNX, qui représente l'échange de réseaux\n"
        "neuronaux ouverts, un standard ouvert pour les formats de modèles d'apprentissage automatique.",

    "module_ondevice_llm_config_response_temperature_label": "Température : {temperature}",
    "module_ondevice_llm_config_response_temperature_input_accessible_description":
        "Ajuste l'aléatoire des réponses du modèle. Des valeurs plus élevées produisent des sorties plus variées,\n"
        "tandis que des valeurs plus basses rendent les réponses plus prévisibles.",

    "module_ondevice_llm_config_response_max_tokens_label": "Nombre maximum de tokens de réponse",
    "module_ondevice_llm_config_response_max_tokens_input_accessible_description":
        "Définit le nombre maximum de tokens à recevoir en réponse, tels que les mots et la ponctuation,\n"
        "contrôlant la longueur de la sortie.",

    "module_ondevice_llm_config_prompt_history_size_label": "Taille de l'historique des prompts",
    "module_ondevice_llm_config_prompt_history_size_input_accessible_description":
        "Contrôle le nombre d'entrées dans l'historique des prompts que le système conserve pour référence.\n"
        "Une valeur zéro permet un nombre illimité d'entrées."
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Enum class for standard and extended color names.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from .enum_base import EnumBase

assert hasattr(EnumBase, 'default'), "Check default() method is implemented in the base class"


class Colors(EnumBase):

    # Red Spectrum
    CRIMSON = ("Crimson", True)
    DARKRED = ("DarkRed", True)
    FIREBRICK = ("Firebrick", True)
    INDIANRED = ("IndianRed", True)
    LIGHTCORAL = ("LightCoral", True)
    MAROON = ("Maroon", False)
    RED = ("Red", False)
    SALMON = ("Salmon", True)
    DARKSALMON = ("DarkSalmon", True)
    TOMATO = ("Tomato", True)

    # Orange Spectrum
    DARKORANGE = ("DarkOrange", True)
    ORANGE = ("Orange", True)
    CORAL = ("Coral", True)
    ORANGERED = ("OrangeRed", True)

    # Yellow Spectrum
    GOLD = ("Gold", True)
    YELLOW = ("Yellow", False)
    LIGHTGOLDENRODYELLOW = ("LightGoldenrodYellow", True)
    PALEGOLDENROD = ("PaleGoldenrod", True)
    GOLDENROD = ("Goldenrod", True)
    DARKGOLDENROD = ("DarkGoldenrod", True)

    # Green Spectrum
    GREENYELLOW = ("GreenYellow", True)
    CHARTREUSE = ("Chartreuse", True)
    LAWNGREEN = ("LawnGreen", True)
    LIME = ("Lime", False)
    LIMEGREEN = ("LimeGreen", True)
    PALEGREEN = ("PaleGreen", True)
    LIGHTGREEN = ("LightGreen", True)
    MEDIUMSPRINGGREEN = ("MediumSpringGreen", True)
    SPRINGGREEN = ("SpringGreen", True)
    MEDIUMSEAGREEN = ("MediumSeaGreen", True)
    SEAGREEN = ("SeaGreen", True)
    FORESTGREEN = ("ForestGreen", True)
    GREEN = ("Green", False)
    DARKGREEN = ("DarkGreen", True)
    YELLOWGREEN = ("YellowGreen", True)
    OLIVEDRAB = ("OliveDrab", True)
    OLIVE = ("Olive", False)

    # Cyan to Blue Spectrum
    MEDIUMAQUAMARINE = ("MediumAquamarine", True)
    AQUAMARINE = ("Aquamarine", True)
    TURQUOISE = ("Turquoise", True)
    MEDIUMTURQUOISE = ("MediumTurquoise", True)
    DARKTURQUOISE = ("DarkTurquoise", True)
    LIGHTSEAGREEN = ("LightSeaGreen", True)
    CADETBLUE = ("CadetBlue", True)
    DARKCYAN = ("DarkCyan", True)
    TEAL = ("Teal", False)
    CYAN = ("Cyan", True)
    LIGHTCYAN = ("LightCyan", True)
    AQUA = ("Aqua", False)
    PALETURQUOISE = ("PaleTurquoise", True)
    AZURE = ("Azure", True)
    ALICEBLUE = ("AliceBlue", True)

    # Blue Spectrum
    NAVY = ("Navy", False)
    DARKBLUE = ("DarkBlue", True)
    MEDIUMBLUE = ("MediumBlue", True)
    BLUE = ("Blue", False)
    MIDNIGHTBLUE = ("MidnightBlue", True)
    ROYALBLUE = ("RoyalBlue", True)
    STEELBLUE = ("SteelBlue", True)
    DODGERBLUE = ("DodgerBlue", True)
    DEEPSKYBLUE = ("DeepSkyBlue", True)
    CORNFLOWERBLUE = ("CornflowerBlue", True)
    SKYBLUE = ("SkyBlue", True)
    LIGHTSKYBLUE = ("LightSkyBlue", True)
    SLATEBLUE = ("SlateBlue", True)
    DARKSLATEBLUE = ("DarkSlateBlue", True)
    MEDIUMSLATEBLUE = ("MediumSlateBlue", True)

    # Violet and Magenta Spectrum
    LAVENDER = ("Lavender", True)
    BLUEVIOLET = ("BlueViolet", True)
    DARKVIOLET = ("DarkViolet", True)
    PURPLE = ("Purple", False)
    MEDIUMPURPLE = ("MediumPurple", True)
    INDIGO = ("Indigo", True)
    DARKORCHID = ("DarkOrchid", True)
    DARKMAGENTA = ("DarkMagenta", True)
    MAGENTA = ("Magenta", True)
    ORCHID = ("Orchid", True)
    VIOLET = ("Violet", True)
    PLUM = ("Plum", True)
    THISTLE = ("Thistle", True)
    LAVENDERBLUSH = ("LavenderBlush", True)
    MEDIUMVIOLETRED = ("MediumVioletRed", True)
    PALEVIOLETRED = ("PaleVioletRed", True)

    # Neutral and Light Spectrum
    GAINSBORO = ("Gainsboro", True)
    LIGHTGREY = ("LightGrey", True)
    SILVER = ("Silver", False)
    DARKGRAY = ("DarkGray", True)
    GRAY = ("Gray", False)
    DIMGRAY = ("DimGray", True)
    LIGHTSLATEGRAY = ("LightSlateGray", True)
    SLATEGRAY = ("SlateGray", True)
    DARKSLATEGRAY = ("DarkSlateGray", True)
    BLACK = ("Black", False, True)
    WHITE = ("White", False)
    SNOW = ("Snow", True)
    HONEYDEW = ("Honeydew", True)
    MINTCREAM = ("MintCream", True)
    GHOSTWHITE = ("GhostWhite", True)
    WHITESMOKE = ("WhiteSmoke", True)
    SEASHELL = ("Seashell", True)
    BEIGE = ("Beige", True)
    OLDLACE = ("OldLace", True)
    FLORALWHITE = ("FloralWhite", True)
    IVORY = ("Ivory", True)
    ANTIQUEWHITE = ("AntiqueWhite", True)
    LINEN = ("Linen", True)
    MISTYROSE = ("MistyRose", True)
    PAPAYAWHIP = ("PapayaWhip", True)
    BLANCHEDALMOND = ("BlanchedAlmond", True)
    BISQUE = ("Bisque", True)
    PEACHPUFF = ("PeachPuff", True)
    NAVAJOWHITE = ("NavajoWhite", True)
    MOCCASIN = ("Moccasin", True)
    CORNSILK = ("CornSilk", True)
    LEMONCHIFFON = ("LemonChiffon", True)
    KHAKI = ("Khaki", True)
    WHEAT = ("Wheat", True)

    def __init__(self, value, is_extended=False, is_default=False, *args, **kwargs):
        super().__init__(value, is_default=is_default, *args, **kwargs)
        # Additional argument
        self.is_extended = is_extended

    @classmethod
    def extended_members(cls):
        # Extended means extended color palette.
        return [member for member in cls if member.is_extended]

    @classmethod
    def is_extended(cls, member):
        # Extended means extended color palette.
        return member in cls.extended_members()



# Russian lexemes common.py
lexemes = {
    "app_title": "Редактор Notolog",
    "app_title_with_sub": "{app_title} - {sub_title}",

    "tree_filter_accessible_desc": "Поле фильтра файлов",

    "menu_action_copy_file_path": "Копировать путь файла",
    "menu_action_rename": "Переименовать",
    "menu_action_delete": "Удалить",
    "menu_action_delete_completely": "Удалить полностью",
    "menu_action_restore": "Восстановить",
    "menu_action_create_new_dir": "Создать новую директорию",

    "dialog_file_rename_title": "Переименовать файл",
    "dialog_file_rename_field_label": "Введите новое имя файла",
    "dialog_file_rename_button_ok": "Переименовать",
    "dialog_file_rename_warning_exists": "Файл с таким именем уже существует",

    "dialog_file_delete_title": "Удалить файл",
    "dialog_file_delete_text": "Удалить файл \"{file_name}\"?",
    "dialog_file_delete_completely_title": "Окончательно удалить файл",
    "dialog_file_delete_completely_text": "Окончательно удалить файл \"{file_name}\"?",
    "dialog_file_delete_error": "Невозможно удалить файл, произошла ошибка",
    "dialog_file_delete_error_not_found": "Файл не найден",

    "dialog_file_restore_title": "Восстановление файла",
    "dialog_file_restore_text": "Восстановить файл \"{file_name}\"?",
    "dialog_file_restore_error": "Не удалось восстановить файл, произошла ошибка",
    "dialog_file_restore_warning_exists": "Файл с именем {file_name} уже существует",

    "dialog_create_new_dir_title": "Создать новую директорию",
    "dialog_create_new_dir_label": "Имя новой директории",
    "dialog_create_new_dir_input_placeholder_text": "Введите имя директории",
    "dialog_create_new_dir_button_ok": "Создать",
    "dialog_create_new_dir_button_cancel": "Отмена",
    "dialog_create_new_dir_warning_empty_title": "Ошибка имени новой директории",
    "dialog_create_new_dir_warning_empty_text": "Имя директории не может быть пустым",
    "dialog_create_new_dir_warning_too_long_title": "Ошибка имени новой директории",
    "dialog_create_new_dir_warning_too_long_text": "Имя директории слишком длинное; максимально допустимое количество "
                                                   "символов: {symbols}!",
    "dialog_create_new_dir_error_existed": "Директория уже существует",
    "dialog_create_new_dir_error": "Не удается создать директорию. Убедитесь, что директория назначения "
                                   "{base_dir} доступна для записи",

    "dialog_message_box_title": "Сообщение",
    "dialog_message_box_button_ok": "Закрыть",

    "action_new_file_first_line_template_text": "Новый документ",
    "action_open_file_dialog_caption": "Открыть файл",
    "action_save_as_file_dialog_caption": "Сохранить файл",

    "dialog_save_empty_file_title": "Сохранить пустой файл",
    "dialog_save_empty_file_text": "Разрешить сохранение файла без содержимого?",

    "dialog_encrypt_file_title": "Зашифровать файл",
    "dialog_encrypt_file_text": "Зашифровать файл \"{file_name}\"?",
    "encrypt_file_warning_file_is_already_encrypted": "Файл уже зашифрован!",
    "dialog_encrypt_file_rewrite_existing_title": "Перезаписать существующий файл",
    "dialog_encrypt_file_rewrite_existing_text": "Перезаписать существующий файл \"{file_path}\"?",

    "dialog_decrypt_file_title": "Расшифровать файл",
    "dialog_decrypt_file_text": "Расшифровать файл \"{file_name}\"?",
    "decrypt_file_warning_file_is_not_encrypted": "Файл не зашифрован!",
    "dialog_decrypt_file_rewrite_existing_title": "Перезаписать существующий файл",
    "dialog_decrypt_file_rewrite_existing_text": "Перезаписать существующий файл \"{file_path}\"?",

    "dialog_encrypt_new_password_title": "Новый пароль",
    "dialog_encrypt_new_password_label": "Пароль:",
    "dialog_encrypt_new_password_input_placeholder_text": "Введите новый пароль",
    "dialog_encrypt_new_password_hint_label": "Подсказка:",
    "dialog_encrypt_new_password_hint_label_description":
        "Подсказка не зашифрована и может быть прочитана из файла!"
        "\nНе используйте очевидные подсказки, которые можно легко угадать,"
        "\nнапример, дату рождения и т.д. Попробуйте использовать отсылку.",
    "dialog_encrypt_new_password_hint_input_placeholder_text": "Введите подсказку (необязательно)",
    "dialog_encrypt_new_password_button_ok": "ОК",
    "dialog_encrypt_new_password_button_cancel": "Отмена",
    "dialog_encrypt_new_password_warning_empty_title": "Предупреждение",
    "dialog_encrypt_new_password_warning_empty_text": "Поле пароля не может быть пустым!",
    "dialog_encrypt_new_password_warning_too_long_title": "Предупреждение",
    "dialog_encrypt_new_password_warning_too_long_text": "Поле подсказки слишком длинное, максимум {symbols} символов!",

    "dialog_encrypt_password_title": "Введите пароль",
    "dialog_encrypt_password_label": "Пароль:",
    "dialog_encrypt_password_input_placeholder_text": "Введите пароль",
    "dialog_encrypt_password_hint_label": "Подсказка:",
    "dialog_encrypt_password_button_ok": "ОК",
    "dialog_encrypt_password_button_cancel": "Отмена",

    "dialog_encrypt_password_reset_title": "Сброс пароля шифрования",
    "dialog_encrypt_password_reset_text": "Вы уверены, что хотите сбросить текущий пароль шифрования?",
    "dialog_encrypt_password_reset_button_cancel": "Отмена",
    "dialog_encrypt_password_reset_button_yes": "Да",

    "dialog_open_link_title": "Ссылка",
    "dialog_open_link_text": "Открыть ссылку \"{url}\" в браузере?",

    "dialog_reset_settings_title": "Сбросить настройки?",
    "dialog_reset_settings_text":
        "Все сохранённые данные в настройках будут очищены, и приложение будет перезапущено для применения изменений.",

    "dialog_exit_unsaved_title": "Подтвердить Выход",
    "dialog_exit_unsaved_text": "Открытый файл '{file_name}' не может быть сохранен. Продолжить выход?",

    "message_app_config_file_access": "Доступ запрещен при обращении к файлу настроек приложения в {file_path}. "
                                      "Установите корректные разрешения для обеспечения правильной работы.",

    "field_dir_path_dialog_caption": "Выбрать каталог",
    "field_file_path_dialog_caption": "Выбрать файл",

    "load_file_encryption_password_mismatch": "Несоответствие пароля шифрования!",
    "load_file_encryption_password_incorrect": "Неверный пароль шифрования!",
    "load_file_none_content_error": "Невозможно загрузить файл.",

    "action_new_file_error_occurred": "Не удается создать файл; произошла ошибка.\nПроверьте разрешения файловой системы.",
    "save_active_file_error_occurred": "Не удается сохранить файл; произошла ошибка.",

    "expandable_block_default_title": "Больше информации...",
    "expandable_block_open_close_tags_mismatch_warning": "Несоответствие открывающих/закрывающих тегов блока <details>",

    "dialog_color_picker_color_copied_to_the_clipboard": "Форматированный текст был скопирован в буфер обмена",

    "popup_about_title": "Информация о приложении",
    "popup_about_app_name_description": "Python Markdown Редактор",

    "popup_about_version": "Версия",
    "popup_about_license": "Лицензия",
    "popup_about_website": "Веб-сайт",
    "popup_about_repository": "GitHub",
    "popup_about_pypi": "PyPi",
    "popup_about_date": "Дата",

    "update_helper_new_version_is_available": "Доступна новая версия {latest_version} приложения",
    "update_helper_latest_version_installed": "Установлена последняя версия приложения",

    "network_connection_error_empty": "Не удаётся получить информацию о ответе",
    "network_connection_error_connection_or_dns":
        "Хост не найден. Возможно, проблемы с интернет-соединением или DNS.",
    "network_connection_error_connection_refused":
        "Соединение отклонено. Возможно, сервер не работает или есть проблемы с сетью.",
    "network_connection_error_connection_timed_out": "Время ожидания соединения истекло. Возможно, есть проблемы с сетью.",
    "network_connection_error_connection_404_error":
        "Ошибка соединения 404. Запрошенная страница или ресурс не найдены.",
    "network_connection_error_generic_with_status_code": "Запрос не выполнен, код состояния: {status_code}",
}



# Portuguese lexemes common.py
lexemes = {
    # Common
    "module_ondevice_llm_name": "LLM no dispositivo",

    "module_ondevice_llm_model_exception": "Erro de modelo não encontrado: {error_msg}",
    "module_ondevice_llm_task_exception": "Não é possível iniciar o gerador de modelo: {error_msg}",
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Enum class for the app's supported themes.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from .enum_base import EnumBase

assert hasattr(EnumBase, 'default'), "Check default() method is implemented in the base class"


class Themes(EnumBase):

    DEFAULT = ("Default", True)  # The second item in the tuple marks this as the default
    CALLIGRAPHY = "Calligraphy"
    NOIR_DARK = "Noir Dark"
    STRAWBERRY = "Strawberry"
    NOCTURNE = "Nocturne"
    SPOOKY = "Spooky"



# German lexemes settings_dialog.py
lexemes = {
    # Einstellungen
    "window_title": "Einstellungen",

    "button_close": "Schließen",

    "tab_general": "Allgemein",
    "tab_editor_config": "Editor",
    "tab_viewer_config": "Betrachter",
    "tab_ai_config": "KI-Konfiguration",

    "general_app_config_label": "App-Konfiguration",
    "general_app_language_label": "Sprache",
    "general_app_language_combo_placeholder_text": "Sprache wählen",
    "general_app_language_combo_accessible_description": "Sprache der App-Oberfläche",
    "general_app_theme_label": "Thema",
    "general_app_theme_combo_placeholder_text": "Thema wählen",
    "general_app_theme_combo_accessible_description": "Thema der App-Oberfläche",
    "general_app_main_menu_label": "Hauptmenü",
    "general_app_main_menu_checkbox": "Hauptmenü anzeigen",
    "general_app_main_menu_checkbox_accessible_description": "Das Haupt-Dropdown-Menü der App anzeigen",
    "general_app_font_size_label": "Schriftgröße: {size}pt",
    "general_app_font_size_slider_accessible_description": "Globale Schriftgröße der App anpassen",

    "general_statusbar_label": "Statusleiste",
    "general_statusbar_show_global_cursor_position_checkbox": "Globale Cursorposition anzeigen",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Globale Cursorposition in der Statusleiste anzeigen",

    "editor_config_label": "Editor-Konfiguration",
    "editor_config_show_line_numbers_checkbox": "Zeilennummern anzeigen",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Zeilennummern im Editor anzeigen",

    "viewer_config_label": "Viewer-Konfiguration",
    "viewer_config_process_emojis_checkbox": "Text-Emojis in Grafiken umwandeln",
    "viewer_config_process_emojis_checkbox_accessible_description": "Text-Emojis in grafische Darstellungen umwandeln",
    "viewer_config_highlight_todos_checkbox": "TODOs hervorheben",
    "viewer_config_highlight_todos_checkbox_accessible_description": "TODOs im Text hervorheben",
    "viewer_config_open_link_confirmation_checkbox": "Bestätigung vor dem Öffnen von Links erforderlich",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Vor dem Öffnen von Links eine Bestätigung anfordern",
    "viewer_config_save_resources_checkbox": "Externe Bilder automatisch auf Festplatte speichern",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Speichert automatisch Kopien von externen Bildern auf der Festplatte für den Offline-Zugriff.",

    "ai_config_inference_module_label": "Inferenzmodul",
    "ai_config_inference_module_names_combo_label": "Aktives Inferenzmodul",
    "ai_config_inference_module_names_combo_placeholder_text": "Modul wählen",
    "ai_config_inference_module_names_combo_accessible_description":
        "Wählen Sie aus den verfügbaren KI-Inferenzmodulen, um mit dem AI-Assistenten zu arbeiten.\n"
        "Optionen umfassen lokale große Sprachmodelle (LLM) mit Echtzeitverarbeitung oder API-basierte Funktionalitäten.",

    "ai_config_base_label": "Grundlegende Parameter",
    "ai_config_multi_turn_dialogue_checkbox": "Mehrstufiger Dialog mit Gesprächsspeicher",
    "ai_config_multi_turn_dialogue_checkbox_accessible_description":
        "Aktivieren Sie einen mehrstufigen Dialog, der die vorherige Eingabe für den Gesprächsspeicher beibehält.\n"
        "Wenn deaktiviert, beeinflussen nur die neue Nachricht und die Systemaufforderung die Antwort.",
    "ai_config_convert_to_md_checkbox": "Ergebnis in Markdown umwandeln",
    "ai_config_convert_to_md_checkbox_accessible_description":
        "Konvertieren Sie die Ausgabemeldung in das Markdown-Format.",
}



# Hindi lexemes main_menu.py
lexemes = {
    "group_file_label": "फ़ाइल",
    "group_edit_label": "संपादित करें",
    "group_tools_label": "उपकरण",
    "group_help_label": "सहायता",

    "actions_file_label_new_document": "नया दस्तावेज़",
    "actions_file_accessible_name_new_document": "नया दस्तावेज़",
    "actions_file_label_open": "खोलें",
    "actions_file_accessible_name_open": "दस्तावेज़ खोलें",
    "actions_file_label_save": "सहेजें",
    "actions_file_accessible_name_save": "दस्तावेज़ सहेजें",
    "actions_file_label_save_as": "इस रूप में सहेजें",
    "actions_file_accessible_name_save_as": "दस्तावेज़ की प्रतिलिपि सहेजें",
    "actions_file_label_settings": "सेटिंग्स",
    "actions_file_accessible_name_settings": "ऐप सेटिंग्स",
    "actions_file_label_reset_settings": "सेटिंग्स रीसेट करें",
    "actions_file_accessible_name_reset_settings": "सेटिंग्स रीसेट करें",
    "actions_file_label_exit": "बाहर निकलें",
    "actions_file_accessible_name_exit": "ऐप से बाहर निकलें",

    "actions_edit_edit_mode": "संपादन",
    "actions_edit_accessible_name_edit_mode": "फ़ाइल संपादित करें",
    "actions_edit_source_mode": "स्रोत",
    "actions_edit_accessible_name_source_mode": "स्रोत कोड",
    "actions_edit_bold": "बोल्ड",
    "actions_edit_accessible_name_bold": "बोल्ड पाठ स्वरूप",
    "actions_edit_italic": "इटैलिक",
    "actions_edit_accessible_name_italic": "इटैलिक पाठ स्वरूप",
    "actions_edit_underline": "रेखांकित",
    "actions_edit_accessible_name_underline": "रेखांकित पाठ स्वरूप",
    "actions_edit_strikethrough": "कटाई",
    "actions_edit_accessible_name_strikethrough": "कटाई पाठ स्वरूप",
    "actions_edit_blockquote": "ब्लॉककोट",
    "actions_edit_accessible_name_blockquote": "ब्लॉककोट पाठ स्वरूप",

    "actions_tools_ai_assistant": "एआई सहायक",
    "actions_tools_accessible_name_ai_assistant": "एआई सहायक शुरू करें",
    "actions_tools_color_picker": "रंग चयनकर्ता",
    "actions_tools_accessible_name_color_picker": "पाठ का रंग चुनें",

    "actions_help_label_md_syntax": "मार्कडाउन वाक्य रचना",
    "actions_help_accessible_name_md_syntax": "मार्कडाउन वाक्य रचना",
    "actions_help_label_check_for_updates": "अपडेट की जाँच करें",
    "actions_help_accessible_name_check_for_updates": "अपडेट की जाँच करें",
    "actions_help_label_bug_report":  "बग रिपोर्ट",
    "actions_help_accessible_name_bug_report": "बग रिपोर्ट जमा करें",
    "actions_help_label_about": "बारे में",
    "actions_help_accessible_name_about": "बारे में"
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Common dialog class.
- Functionality: Acts as a generic dialog class for standard app dialogues.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QLabel, QVBoxLayout, QDialogButtonBox, QSizePolicy
from PySide6.QtGui import QFontMetrics

import logging


class CommonDialog(QDialog):

    def __init__(self, title=str, text=str, callback=None, reject_callback=None, parent=None):
        super().__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the dialog
            self.setFont(self.parent.font())

        self.logger = logging.getLogger('common_dialog')

        self.setWindowTitle(str(title))
        self.setObjectName('common_dialog')

        # Set dialog size derived from the main window size
        """
        main_window_size = parent.size()
        # Will be adjusted with size hint later at resizeEvent()
        dialog_width = int(main_window_size.width() * 0.33)
        dialog_height = int(main_window_size.height() * 0.33)
        self.resize(dialog_width, dialog_height)
        """
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        if self.sizeHint().isValid():
            self.setMinimumSize(self.sizeHint())

        label = QLabel()
        label.setObjectName('common_label')
        label.setMargin(10)  # Set vertical padding for the dialog label
        label.setText(str(text))
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        label.adjustSize()
        """
        Other examples of the widget geometry's adjustments:
        label.move(128, 64)
        label.setGeometry(QStyle.alignedRect(
            Qt.LayoutDirection.LeftToRight,
            Qt.AlignmentFlag.AlignCenter,
            dialog.size(),
            dialog.geometry()))
        """

        button_box = QDialogButtonBox()
        button_box.setObjectName('button_box')
        button_box.setOrientation(Qt.Orientation.Horizontal)
        button_box.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Yes)
        button_box.setCenterButtons(True)

        # not QVBoxLayout(self) to avoid: 'Attempting to add QLayout "" to ... "", which already has a layout
        layout = QVBoxLayout()
        layout.addWidget(label)
        layout.addWidget(button_box)
        self.setLayout(layout)

        self.adjust_minimum_width(str(title))

        if callable(callback):
            button_box.accepted.connect(lambda: callback(self.reject))
        else:
            self.logger.debug('No callback method was provided for the dialog!')
            button_box.accepted.connect(self.close)

        if callable(reject_callback):
            button_box.rejected.connect(lambda: reject_callback(self.reject))
        else:
            self.logger.debug('No reject callback method was provided for the dialog!')
            button_box.rejected.connect(self.reject)

    def resizeEvent(self, event) -> None:
        """
        To make dialog stretch if the content size is longer than expected
        @param event: QEvent
        @return: None
        """
        # Get the preferred size of the dialog's content
        preferred_size = self.layout().sizeHint()
        # Adjust the dialog's size based on the preferred size then
        if preferred_size.isValid():
            self.resize(preferred_size)

    def adjust_minimum_width(self, title: str) -> None:
        # Calculate width required for the title text
        font_metrics = QFontMetrics(self.font())
        title_width = font_metrics.horizontalAdvance(title)

        # Set minimum width for the dialog
        self.setMinimumWidth(title_width + 150)  # Add some padding



# Georgian lexemes main_menu.py
lexemes = {
    "group_file_label": "ფაილი",
    "group_edit_label": "რედაქტირება",
    "group_tools_label": "ინსტრუმენტები",
    "group_help_label": "დახმარება",

    "actions_file_label_new_document": "ახალი დოკუმენტი",
    "actions_file_accessible_name_new_document": "ახალი დოკუმენტი",
    "actions_file_label_open": "გახსნა",
    "actions_file_accessible_name_open": "დოკუმენტის გახსნა",
    "actions_file_label_save": "შენახვა",
    "actions_file_accessible_name_save": "დოკუმენტის შენახვა",
    "actions_file_label_save_as": "შენახვა როგორც",
    "actions_file_accessible_name_save_as": "დოკუმენტის ასლის შენახვა",
    "actions_file_label_settings": "პარამეტრები",
    "actions_file_accessible_name_settings": "აპლიკაციის პარამეტრები",
    "actions_file_label_reset_settings": "პარამეტრების აღდგენა",
    "actions_file_accessible_name_reset_settings": "პარამეტრების აღდგენა",
    "actions_file_label_exit": "გასვლა",
    "actions_file_accessible_name_exit": "აპლიკაციის დატოვება",

    "actions_edit_edit_mode": "რედაქტირების რეჟიმი",
    "actions_edit_accessible_name_edit_mode": "ფაილის რედაქტირება",
    "actions_edit_source_mode": "წყარო",
    "actions_edit_accessible_name_source_mode": "წყაროს კოდი",
    "actions_edit_bold": "მუქი",
    "actions_edit_accessible_name_bold": "ტექსტის მუქი ფორმატი",
    "actions_edit_italic": "კურსივი",
    "actions_edit_accessible_name_italic": "ტექსტის კურსივის ფორმატი",
    "actions_edit_underline": "ხაზგასმა",
    "actions_edit_accessible_name_underline": "ტექსტის ხაზგასმის ფორმატი",
    "actions_edit_strikethrough": "ხაზზეთმის ფორმატი",
    "actions_edit_accessible_name_strikethrough": "ტექსტის ხაზზეთმის ფორმატი",
    "actions_edit_blockquote": "ციტატა",
    "actions_edit_accessible_name_blockquote": "ტექსტის ციტირების ფორმატი",

    "actions_tools_ai_assistant": "AI ასისტენტი",
    "actions_tools_accessible_name_ai_assistant": "AI ასისტენტის გაშვება",
    "actions_tools_color_picker": "ფერის შერჩევის ინსტრუმენტი",
    "actions_tools_accessible_name_color_picker": "ტექსტის ფერის არჩევა",

    "actions_help_label_md_syntax": "Markdown სინტაქსი",
    "actions_help_accessible_name_md_syntax": "Markdown სინტაქსი",
    "actions_help_label_check_for_updates": "განახლებების შემოწმება",
    "actions_help_accessible_name_check_for_updates": "განახლებების შემოწმება",
    "actions_help_label_bug_report": "შეცდომის რეპორტი",
    "actions_help_accessible_name_bug_report": "შეცდომის რეპორტის შეყვანა",
    "actions_help_label_about": "შესახებ",
    "actions_help_accessible_name_about": "აპლიკაციის შესახებ",
}



# Greek lexemes settings_dialog.py
lexemes = {
    "tab_module_llama_cpp_config": "Μονάδα llama.cpp",

    "module_llama_cpp_config_label": "Μονάδα llama.cpp",
    "module_llama_cpp_config_path_label": "Τοποθεσία Μοντέλου",
    "module_llama_cpp_config_path_input_placeholder_text": "Επιλέξτε ή εισάγετε τη διαδρομή του μοντέλου",
    "module_llama_cpp_config_path_input_accessible_description":
        "Ένα πεδίο εισόδου με επιλογέα για να καθορίσετε την τοποθεσία του τοπικού μοντέλου.\n"
        "Υποστηρίζει μοντέλα σε μορφή GGUF, μια δυαδική μορφή αρχείου βελτιστοποιημένη για την\n"
        "αποθήκευση μοντέλων που χρησιμοποιούνται με GGML και εκτελεστές βασισμένους σε GGML.",
    "module_llama_cpp_config_path_input_filter_text": "Αρχεία GGUF",

    "module_llama_cpp_config_context_window_label": "Μέγεθος Παραθύρου Πλαισίου",
    "module_llama_cpp_config_context_window_input_accessible_description":
        "Καθορίζει τον αριθμό των tokens που λαμβάνει υπόψη το μοντέλο για τη δημιουργία απαντήσεων.\n"
        "Ελέγχει πόσο προηγούμενο πλαίσιο χρησιμοποιείται.",

    "module_llama_cpp_chat_formats_label": "Μορφές Συνομιλίας",
    "module_llama_cpp_chat_formats_combo_placeholder_text": "Επιλέξτε μια μορφή συνομιλίας",
    "module_llama_cpp_chat_formats_combo_accessible_description":
        "Μενού dropdown για την επιλογή της μορφής που χρησιμοποιείται για τις συνομιλίες μοντέλων.",

    "module_llama_cpp_config_system_prompt_label": "Προτροπή Συστήματος",
    "module_llama_cpp_config_system_prompt_edit_placeholder_text": "Εισάγετε κείμενο προτροπής συστήματος",
    "module_llama_cpp_config_system_prompt_edit_accessible_description":
        "Πεδίο κειμένου για την εισαγωγή προτροπών συστήματος που καθοδηγούν τις απαντήσεις του μοντέλου.",

    "module_llama_cpp_config_response_temperature_label": "Θερμοκρασία Απάντησης: {temperature}",
    "module_llama_cpp_config_response_temperature_input_accessible_description":
        "Ρυθμίζει την τυχαιότητα των απαντήσεων του μοντέλου. Υψηλότερες τιμές παράγουν πιο ποικίλες εξόδους,\n"
        "ενώ χαμηλότερες τιμές οδηγούν σε πιο προβλέψιμες απαντήσεις.",

    "module_llama_cpp_config_response_max_tokens_label": "Μέγιστοι Tokens ανά Απάντηση",
    "module_llama_cpp_config_response_max_tokens_input_accessible_description":
        "Περιορίζει τον αριθμό των tokens στις απαντήσεις του μοντέλου μέχρι το πραγματικό όριο του παραθύρου πλαισίου.\n"
        "Μηδενική τιμή υποθέτει τη χωρητικότητα του παραθύρου πλαισίου.",

    "module_llama_cpp_config_prompt_history_size_label": "Μέγεθος Ιστορικού Προτροπών",
    "module_llama_cpp_config_prompt_history_size_input_accessible_description":
        "Ελέγχει τον αριθμό των καταχωρήσεων στο ιστορικό προτροπών που διατηρούνται από το σύστημα για αναφορά.\n"
        "Μηδενική τιμή επιτρέπει απεριόριστες καταχωρήσεις."
}



# Latin lexemes settings_dialog.py
lexemes = {
    "tab_module_llama_cpp_config": "Modulus llama.cpp",

    "module_llama_cpp_config_label": "Modulus llama.cpp",
    "module_llama_cpp_config_path_label": "Locatio Modeli",
    "module_llama_cpp_config_path_input_placeholder_text": "Elige aut inscribe viam modeli",
    "module_llama_cpp_config_path_input_accessible_description":
        "Campus input cum selectorio ad viam modeli localis specificandam. Sustinet modela in forma GGUF,\n"
        "formatum file binarium optimatum ad reponendum modela usui cum GGML et executoribus GGML-fundatis.",
    "module_llama_cpp_config_path_input_filter_text": "GGUF Files",

    "module_llama_cpp_config_context_window_label": "Magnitudo Fenestrae Contextus",
    "module_llama_cpp_config_context_window_input_accessible_description":
        "Statuit numerum signorum quae model considerat ad responsa generanda. Moderatur quantum contextus prioris utitur.",

    "module_llama_cpp_chat_formats_label": "Formae Colloquii",
    "module_llama_cpp_chat_formats_combo_placeholder_text": "Elige formam colloquii",
    "module_llama_cpp_chat_formats_combo_accessible_description":
        "Menu dropdown ad formam eligendam quae ad colloquia modeli utitur.",

    "module_llama_cpp_config_system_prompt_label": "Promptum Systematis",
    "module_llama_cpp_config_system_prompt_edit_placeholder_text": "Inscribe textum prompti systematis",
    "module_llama_cpp_config_system_prompt_edit_accessible_description":
        "Campus texti ad prompta systematis ingredienda quae responsa modeli dirigunt.",

    "module_llama_cpp_config_response_temperature_label": "Temperatura Responsi: {temperature}",
    "module_llama_cpp_config_response_temperature_input_accessible_description":
        "Adiustat casualitatem responsorum modeli. Maiores valores producunt outputs variabilius,\n"
        "dum valores inferiores resultunt in responsis praedicibilioribus.",

    "module_llama_cpp_config_response_max_tokens_label": "Maxima Signorum per Responsionem",
    "module_llama_cpp_config_response_max_tokens_input_accessible_description":
        "Limitat numerum signorum in responsis modeli ad limitem fenestrae contextus actualis.\n"
        "Valor nihili capacitas fenestrae contextus assumit.",

    "module_llama_cpp_config_prompt_history_size_label": "Magnitudo Historiae Promptorum",
    "module_llama_cpp_config_prompt_history_size_input_accessible_description":
        "Moderatur numerum invectionum in historia promptorum quae a systemate retentae sunt ad referentiam.\n"
        "Valor nihili invectiones illimitatas permittit."
}



from PySide6.QtCore import QObject, QStandardPaths, Signal

import os
import tomli
import tomli_w
import logging
import shutil

from notolog.app_package import AppPackage

from typing import Union
from threading import Lock

toml_base_app_config = """
[app]
name = "Notolog"
version = "1.1.1"
license = "MIT License"
date = "2024-2025"
website = "https://notolog.app"
repository = "https://github.com/notolog/notolog-editor"
pypi = "https://pypi.org/project/notolog"
author = "Vadim Bakhrenkov"

[repository]

    [repository.github]
    username = "notolog"
    project = "notolog-editor"
    api_release_url = "https://api.github.com/repos/notolog/notolog-editor/releases/latest"
    release_url = "https://github.com/notolog/notolog-editor/releases/latest"
    bug_report_url = "https://github.com/notolog/notolog-editor/issues"

[settings]
org_name = "Notolog"
org_domain = "notolog.app"
app_name = "notolog_editor"
"""

toml_app_config_header = """# Notolog Editor
# An open-source Markdown editor built with Python.

# Application-Level Configuration
# --------------------------------
# This section contains default configuration options for the application.
# These settings can be overridden programmatically at runtime.
#
# IMPORTANT:
# This file is auto-generated. Changes made to this file may be overwritten.
# To customize configuration safely, ensure modifications are done through
# the application's interface or according to the documented process.
#
# Modifying this file directly is not recommended as it could lead to
# unexpected behavior or loss of your settings upon the next auto-generation.\n
"""

# This is a fallback option if app_config.toml is not found
toml_app_config = """
[font]
base_size = 12
min_size = 5
max_size = 42

[logger]
level = "info"

[security]
app_secret = ""

[editor]
media_dir = "images"

[package]
type = ""
"""


class AppConfig(QObject):
    """
    App configuration to share between app's objects.
    """

    _instance = None  # Singleton instance
    _lock = Lock()

    value_changed = Signal(dict)  # type: Signal[dict]

    default_package = AppPackage().default_package

    default_logger_level = logging.INFO

    def __new__(cls, *args, **kwargs):
        # Overriding __new__ to control the instantiation process
        if not cls._instance:
            with cls._lock:
                # Create the instance if it doesn't exist
                cls._instance = super(AppConfig, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        # Prevent re-initialization if the instance is already set up.
        if hasattr(self, 'base_app_config'):
            return

        # Ensure that the initialization check and the setting of base_app_config are atomic.
        with self._lock:
            # Double-check to prevent race conditions during initialization.
            if hasattr(self, 'base_app_config'):
                return

            # Initialize the QObject part only once
            super(AppConfig, self).__init__()

            self.logger = logging.getLogger('app_config')

            # Config variables
            self.base_app_config = None  # type: Union[dict, None]
            self.toml_file_path = None  # type: Union[str, None]
            self.app_config = None  # type: Union[dict, None]

            # Font settings
            self._font_size = None  # Base value
            self._prev_font_size = None  # Previous value

            # For pytest to allow override some params
            self._test_mode = False

            # Initialize
            self.load_initial_conf()

            self.logger.info('App config is activated')

            self.value_changed.connect(self.app_config_update_handler)

    @classmethod
    def get_instance(cls):
        # Class method to get the singleton instance
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def get_app_config_path(self):
        """
        Method to get app config file path which is depends on the app package type stored in the 'package' file.
        The method can be called before the actual package type if the 'package' file yet not exists in some cases.

        Package file purpose to distinguish the actual installation whither through pip, deb or similar packages.
        This approach allows to keep app config and it's settings separately from each other.

        returns: str path to the config file stored in a preferred system location (QStandardPaths.ApplicationsLocation).
        """

        # Get the standard configuration path
        config_dir = QStandardPaths.writableLocation(QStandardPaths.StandardLocation.ApplicationsLocation)

        try:
            # Ensure the config directory exists
            os.makedirs(config_dir, exist_ok=True)
        except OSError as e:
            # Handle specific errors, e.g., permission errors
            self.logger.warning(f"Error: Unable to create directory '{config_dir}'; error occurred {e}")

        # Suffix for testing to avoid modifying the real configuration
        file_name_suffix = '_qa' if self.get_test_mode() else ''

        # Add package type suffix if it's not the default
        app_package = self.get_package_type()
        if app_package != self.default_package:
            file_name_suffix += f'_{app_package}'

        # Get complete app_config.toml file path
        config_path = os.path.join(
            config_dir,
            f'notolog_app_config{file_name_suffix}.toml'
        )

        # For compatibility with previous versions, clone config from the previous location
        if not os.path.exists(config_path):
            # Get this file dir path
            script_dir = os.path.dirname(os.path.realpath(__file__))
            prev_conf_path = os.path.join(os.path.dirname(script_dir), 'app_config.toml')
            if os.path.exists(prev_conf_path):
                # Copy previous version of the config to the new one
                shutil.copy(prev_conf_path, config_path)
                # Delete previous file
                os.remove(prev_conf_path)

        return config_path

    @staticmethod
    def get_base_app_config():
        return tomli.loads(toml_base_app_config)

    @staticmethod
    def get_default_app_config():
        return tomli.loads(toml_app_config)

    def get_config(self):
        return self.app_config

    def load_config(self, config_file_path) -> Union[dict, None]:
        if not (os.path.exists(config_file_path) and os.access(config_file_path, os.R_OK)):
            return None
        try:
            # Read from the actual app_config.toml file
            with open(config_file_path, 'r') as config_file:
                return tomli.loads(config_file.read())
        except FileNotFoundError:
            self.logger.warning(f"The configuration file '{config_file_path}' was not found.")
        except PermissionError:
            self.logger.warning(f"Permission denied while accessing the file '{config_file_path}'.")
        except tomli.TOMLDecodeError:
            self.logger.warning(f"Error decoding TOML from the file '{config_file_path}'.")
        except Exception as e:
            self.logger.warning(f"An unexpected error occurred while loading the config: {e}")
        return None

    def load_initial_conf(self):
        # Clear stored data if set
        self.app_config = {}

        # Get base config
        self.base_app_config = self.get_base_app_config()

        # Load the default configuration schema from a TOML string
        self.app_config = self.get_default_app_config()

        # Load package config to update the configuration file if needed
        app_config_package = AppPackage().get_config()
        if not AppPackage().validate_config(app_config_package):
            # self.logger.warning(f"Package config contains unsupported data '{str(app_config_package)[:255]}' or missing.")
            pass
        else:
            # Merge the existing package config into the default one
            self.app_config.update(app_config_package)

        # Get the current app config file path (it might have just been updated)
        self.toml_file_path = self.get_app_config_path()

        # In case the config file is missing
        if not os.path.exists(self.toml_file_path) and os.access(os.path.dirname(self.toml_file_path), os.W_OK):
            # Create the config.
            self.save_app_config()

        stored_app_config = self.load_config(self.toml_file_path)
        # Check if existing config is available
        if self.validate_config(stored_app_config):
            # Merge the existing app config into the default one
            self.app_config.update(stored_app_config)
            if ('package' in self.app_config
                    and 'package' in app_config_package
                    and self.app_config['package'] != app_config_package['package']):
                # Such discrepancy can possibly occur for various reasons
                self.logger.warning(f"The configuration file package {self.app_config['package']} "
                                    f"and {app_config_package['package']} are differ.")
                # Merge the package config into the combined one again to restore
                self.app_config.update(app_config_package)

        # Font settings
        self._font_size = self.app_config['font']['base_size']  # Base value
        self._prev_font_size = self.app_config['font']['base_size']  # Previous value

    def delete_app_config(self) -> bool:
        """
        Attempts to delete configuration file if it exists and is writable.

        Returns:
            bool: True if the file was successfully deleted, False otherwise.
        """

        # Check if file exists and is writable
        if os.path.exists(self.toml_file_path) and os.access(self.toml_file_path, os.W_OK):
            try:
                os.remove(self.toml_file_path)
                self.toml_file_path = None
                return True
            except OSError as e:
                self.logger.warning(f"Error deleting configuration file {self.toml_file_path}: {e}")
                return False
        else:
            self.logger.warning(f"Configuration file does not exist or is not writable: {self.toml_file_path}")
            return False

    def save_app_config(self):
        try:
            # Regenerate the app_config.toml file with new settings
            with open(self.toml_file_path, 'wb') as f:
                f.write(toml_app_config_header.encode('utf-8'))
                tomli_w.dump(self.app_config, f)
        except PermissionError:
            self.logger.warning(f"Permission denied: Cannot write to the file {self.toml_file_path}.")
        except IOError as e:
            self.logger.warning(f"Failed to write config to file {self.toml_file_path}: {e}")
        except Exception as e:
            self.logger.warning(f"An unexpected error occurred: {e}")

    def app_config_update_handler(self, data: dict) -> None:
        # Data contains information about the changes made
        self.logger.debug('App config handler is in use "%s"' % data)

        # Save the current config
        self.save_app_config()

    def setup_package(self, app_package) -> None:
        # Validate package
        if not AppPackage().validate_type(app_package):
            self.logger.warning(
                f"Trying to setup unsupported package '{str(app_package)[:16]}'; "
                f"fallback to the default package '{self.default_package}'")
            app_package = self.default_package

        # Save default config
        AppPackage().set_type(app_package)

        # Reinitialize the updated config and its path
        self.load_initial_conf()

        # Block signals
        was_blocked = self.blockSignals(True)

        # Verify the initial setup and update the config
        self.save_app_config()

        # Restore signals
        self.blockSignals(was_blocked)

    def validate_config(self, app_config):
        return all(key in app_config for key in self.get_default_app_config())

    """
    Constant values with getters.
    """

    def get_app_name(self) -> str:
        return self.base_app_config['app']['name']

    def get_app_version(self) -> str:
        return self.base_app_config['app']['version']

    def get_app_license(self) -> str:
        return self.base_app_config['app']['license']

    def get_app_date(self) -> str:
        return self.base_app_config['app']['date']

    def get_app_website(self) -> str:
        return self.base_app_config['app']['website']

    def get_app_repository(self) -> str:
        return self.base_app_config['app']['repository']

    def get_app_pypi(self) -> str:
        return self.base_app_config['app']['pypi']

    def get_app_author(self) -> str:
        return self.base_app_config['app']['author']

    def get_repository_github_username(self) -> str:
        return self.base_app_config['repository']['github']['username']

    def get_repository_github_project(self) -> str:
        return self.base_app_config['repository']['github']['username']

    def get_repository_github_api_release_url(self) -> str:
        return self.base_app_config['repository']['github']['api_release_url']

    def get_repository_github_release_url(self) -> str:
        return self.base_app_config['repository']['github']['release_url']

    def get_repository_github_bug_report_url(self) -> str:
        return self.base_app_config['repository']['github']['bug_report_url']

    def get_settings_org_name(self) -> str:
        return self.base_app_config['settings']['org_name']

    def get_settings_org_domain(self) -> str:
        return self.base_app_config['settings']['org_domain']

    def get_settings_app_name(self) -> str:
        return self.base_app_config['settings']['app_name']

    def get_settings_app_name_qa(self) -> str:
        return f"{self.base_app_config['settings']['app_name']}_qa"

    def get_font_base_size(self) -> int:
        return self.app_config['font']['base_size']

    def get_editor_media_dir(self) -> str:
        return self.app_config['editor']['media_dir']

    """
    Methods with setter to override default value.
    """

    def set_logger_level(self, value) -> None:
        self.app_config['logger']['level'] = value
        self.value_changed.emit({'logger_level': value})

    def get_logger_level(self) -> int:
        level_name = None
        if self.app_config and 'logger' in self.app_config and 'level' in self.app_config['logger']:
            level_name = (self.app_config['logger']['level']).upper()

        try:
            # Use the modern method available in Python 3.11+
            available_levels = logging.getLevelNamesMapping()
        except AttributeError:
            # Fallback for earlier Python versions
            if hasattr(logging, '_nameToLevel'):
                available_levels = logging._nameToLevel
            else:
                available_levels = {}

        if level_name not in available_levels:
            self.logger.warning(f"Logger level is not recognized: '{level_name}'.\n\t"
                                f"Available options are: {str(available_levels.keys())}")
            return self.default_logger_level

        # Return the level
        return available_levels[level_name]

    def set_font_min_size(self, value) -> None:
        self.app_config['font']['min_size'] = value
        self.value_changed.emit({'font_min_size': value})

    def get_font_min_size(self) -> int:
        return self.app_config['font']['min_size']

    def set_font_max_size(self, value) -> None:
        self.app_config['font']['max_size'] = value
        self.value_changed.emit({'font_max_size': value})

    def get_font_max_size(self) -> int:
        return self.app_config['font']['max_size']

    def set_security_app_secret(self, value) -> None:
        self.app_config['security']['app_secret'] = value
        self.value_changed.emit({'security_app_secret': value})

    def get_security_app_secret(self) -> str:
        return self.app_config['security']['app_secret']

    def set_package_type(self, value) -> None:
        self.app_config['package']['type'] = value
        self.value_changed.emit({'package_type': value})

    def get_package_type(self) -> str:
        """
        Retrieves the package type specified in the app's configuration.
        If no package is found, the method returns `self.default_package` as a fallback.

        Returns:
            str: The package name or `self.default_package` if no package is found.
        """

        package = None
        if self.app_config and 'package' in self.app_config and 'type' in self.app_config['package']:
            package = self.app_config['package']['type']

        # Return the package
        return package if package else self.default_package

    """
    Class system variables.
    Do not emit signals upon their update.
    """

    def set_font_size(self, value) -> None:
        self._font_size = value

    def get_font_size(self) -> int:
        return self._font_size

    def set_prev_font_size(self, value) -> None:
        self._prev_font_size = value

    def get_prev_font_size(self) -> int:
        return self._prev_font_size

    def set_test_mode(self, value) -> None:
        self._test_mode = value

    def get_test_mode(self) -> bool:
        return self._test_mode



# Turkish lexemes settings_dialog.py
lexemes = {
    "tab_ondevice_llm_config": "Cihazda LLM",

    "module_ondevice_llm_config_label": "Cihazda LLM Modeli",
    "module_ondevice_llm_config_path_label": "ONNX Model Konumu",
    "module_ondevice_llm_config_path_input_placeholder_text": "Model dizinine yol",
    "module_ondevice_llm_config_path_input_accessible_description":
        "ONNX dosyalarının bulunduğu model dizinine yol belirtmek için bir seçiciye sahip giriş alanı.\n"
        "Desteklenen modeller ONNX formatındadır, bu da Açık Sinir Ağı Değişimini temsil eder, açık standart\n"
        "makine öğrenimi modelleri formatıdır.",

    "module_ondevice_llm_config_response_temperature_label": "Sıcaklık: {temperature}",
    "module_ondevice_llm_config_response_temperature_input_accessible_description":
        "Modelin yanıtlarının rastgeleliğini ayarlar. Daha yüksek değerler daha çeşitli çıktılar üretir,\n"
        "daha düşük değerler ise yanıtları daha tahmin edilebilir kılar.",

    "module_ondevice_llm_config_response_max_tokens_label": "Maksimum Yanıt Token'ları",
    "module_ondevice_llm_config_response_max_tokens_input_accessible_description":
        "Bir yanıtta alınabilecek maksimum token sayısını belirler, bu da kelimeler ve noktalama işaretlerini içerir,\n"
        "çıktının uzunluğunu kontrol eder.",

    "module_ondevice_llm_config_prompt_history_size_label": "Komut Geçmişi Boyutu",
    "module_ondevice_llm_config_prompt_history_size_input_accessible_description":
        "Sistem tarafından referans için tutulan komut geçmişindeki giriş sayısını kontrol eder.\n"
        "Sıfır değer sınırsız girişe izin verir."
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Async image downloader.
- Functionality: Automatically downloads image resources mentioned in the markdown document and caches them in memory
  or on the drive.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Signal, QObject, QUrl, QDir, QByteArray
from PySide6.QtGui import QPixmap, QPixmapCache
from PySide6.QtNetwork import QNetworkAccessManager, QNetworkRequest, QNetworkReply

from .settings import Settings
from .app_config import AppConfig
from .lexemes.lexemes import Lexemes
from .helpers.file_helper import size_f, save_file

from urllib.parse import urlparse
from typing import Union

from qasync import asyncSlot
import asyncio

import os
import re
import hashlib
import logging


class ImageDownloader(QObject):  # QObject to allow signal emitting

    # Signal to emit upon single resource was downloaded
    downloaded = Signal(dict)

    # Signal to emit upon all downloading tasks in queue have been finished
    finished = Signal(int)

    RESOURCE_DIR = AppConfig().get_editor_media_dir()

    event_loop = None  # async event loop
    resource_tasks: list = []  # async downloads task queue
    folder: Union[QDir, None] = None  # active folder to work with resources
    base_folder: str = None  # base folder to work with
    downloaded_cnt: int = 0  # downloads counter, counts till finished emitted
    network_manager: QNetworkAccessManager = None

    def __init__(self, base_folder: str = None):
        super().__init__()

        self.settings = Settings(parent=self)

        self.logger = logging.getLogger('image_downloader')

        self.lexemes = Lexemes()

        self.base_folder = base_folder

        # Setup start up params
        self.initialize()

    def initialize(self):
        self.update_resource_folder(self.base_folder)
        self.network_manager = QNetworkAccessManager(self)

    def get_resource_folder(self, base_folder: str = None) -> QDir:
        # Check folder and make it QDir
        if base_folder:
            folder = os.path.join(base_folder, self.RESOURCE_DIR)
        else:
            folder = os.path.join(QDir.currentPath(), self.RESOURCE_DIR)
        if isinstance(folder, str):
            folder = QDir(folder)
        return folder

    def update_resource_folder(self, base_folder: str = None) -> None:
        # Validate folder and make it QDir
        self.folder = self.get_resource_folder(base_folder)

        self.logger.debug(f'Folder to save downloaded resources: {self.folder.path()}')

    def download_image(self, url: str) -> None:
        if not self.is_external_url(url):
            self.logger.debug(f"Skip downloading as the provided url is not an external url: {url}")
            return

        self.logger.debug(f'Downloading resource from "{url}"')

        request = QNetworkRequest(QUrl(url))
        reply = self.network_manager.get(request)

        # Connecting finished signal
        reply.finished.connect(lambda _reply=reply: self.handle_network_reply(_reply))
        # Connecting error signal
        reply.errorOccurred.connect(lambda error, _reply=reply: self.handle_network_reply_error(_reply, error))

    def save_image(self, url: str, data: QByteArray) -> None:
        # file_name = os.path.basename(url)
        file_name = self.url_to_filename(url)
        # Alternative approach:
        # file_extension = self.mime_to_extension(mime_type)
        # file_name = f"{file_name}{file_extension}"
        # Create resource folder if not yet created
        if os.access(os.path.dirname(self.folder.path()), os.W_OK):
            os.makedirs(self.folder.path(), exist_ok=True)
        else:
            self.logger.warning(f"Permission denied: Cannot create directory {self.folder.path()}")
            return
        # File path to save
        file_path = os.path.join(self.folder.path(), file_name)
        # Save received data
        if save_file(file_path, data, as_bytearray=True):
            self.logger.debug(f"Resource saved {url} to {file_path} [{size_f(len(data))}]")
        else:
            self.logger.debug(f"Resource not saved {url} to {file_path} [{size_f(len(data))}]")

    def handle_network_reply(self, reply):
        # Get received status code, say 200
        status_code = reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute)

        self.logger.debug(f'Network reply, reply: {reply}, status_code: {status_code}')

        # Make sure there is no error
        if reply.error() == QNetworkReply.NetworkError.NoError:
            # No error, processing the image
            result_message = self.lexemes.get('network_connection_error_empty')
            self.logger.debug(result_message)

            # Check if the resource is an image
            mime_type = reply.header(QNetworkRequest.KnownHeaders.ContentTypeHeader)
            if 'image' in mime_type:
                url = reply.url().toString()
                data = reply.readAll()  # type: QByteArray

                self.logger.debug(f"Resource data downloaded {url} [{size_f(len(data))}]")

                # Store the image within the app's cache first
                self.cache_pixmap(url, data)

                # Save file if settings allow
                if self.settings.viewer_save_resources:
                    # Saving downloaded files
                    self.save_image(url, data)

                # Emit the file downloaded signal
                self.downloaded.emit({'resource_name': url})
                self.downloaded_cnt += 1

        reply.finished.disconnect()
        reply.errorOccurred.disconnect()
        reply.deleteLater()  # Clean up the QNetworkReply object

    def cache_pixmap(self, url: str, data: QByteArray):
        # Store the image within the app's cache first
        pixmap = QPixmap()
        pixmap.loadFromData(data)
        QPixmapCache.insert(url, pixmap)

    def handle_network_reply_error(self, reply, error_code=None):
        # Get received status code, say 404
        status_code = reply.attribute(QNetworkRequest.Attribute.HttpStatusCodeAttribute)

        self.logger.debug(f'Network reply, reply: {reply}, error_code: {error_code}, status_code: {status_code}')

        if reply.error() == QNetworkReply.NetworkError.HostNotFoundError:
            # The host was not found, indicating possible DNS issues or no internet connection
            result_message = self.lexemes.get('network_connection_error_connection_or_dns')
        elif reply.error() == QNetworkReply.NetworkError.ConnectionRefusedError:
            # Connection was refused, indicating the server might be down or there are network issues
            result_message = self.lexemes.get('network_connection_error_connection_refused')
        elif reply.error() == QNetworkReply.NetworkError.TimeoutError:
            # The connection timed out, indicating network issues
            result_message = self.lexemes.get('network_connection_error_connection_timed_out')
        elif reply.error() == QNetworkReply.NetworkError.ContentNotFoundError:
            # The requested page or resource is not found
            result_message = self.lexemes.get('network_connection_error_connection_404_error')
        else:
            # Handle other errors
            result_message = self.lexemes.get('network_connection_error_generic_with_status_code',
                                              status_code=status_code)

        self.logger.warning(result_message)
        self.logger.warning(f'Failed to download resource "{reply.errorString()}" with error code [{error_code}]')

        reply.finished.disconnect()
        reply.errorOccurred.disconnect()
        reply.deleteLater()  # Clean up the QNetworkReply object

    @asyncSlot()
    async def download_resource_in_queue(self, resource_url) -> None:
        # Check async loop is running
        if not asyncio.get_event_loop().is_running():
            self.logger.debug('Skipping the task because the async loop is not running.')
            return

        self.logger.debug('Downloading resource %d tasks in queue' % len(self.resource_tasks))

        task = asyncio.ensure_future(self.resource_download_async(resource_url))  # task: .download_resource_in_queue()
        # Add task to the local pool first
        self.resource_tasks.append(task)
        # Callback method to set up further actions
        task.add_done_callback(lambda _task: self.resource_task_callback(_task))  # task: .resource_download_async()

        try:
            done, pending = await asyncio.wait(
                self.resource_tasks,
                return_when=asyncio.ALL_COMPLETED,  # There is no pending tasks check
                # timeout = 1.5  # to return after the timeout, some task could be pending
            )
            self.logger.debug(f'Downloading resource tasks progress. Done {len(done)}, pending {len(pending)}')
        except asyncio.CancelledError:
            # All tasks will be cancelled later upon close event
            pass

        if len(self.resource_tasks) <= 1:
            # Hold this method a particular amount of time to allow completed tasks to settle as this step is
            # typically ahead.
            try:
                await asyncio.sleep(0.25, self.event_loop)
            except asyncio.CancelledError:
                # All tasks will be cancelled later upon close event
                pass
            if self.downloaded_cnt > 0:
                # All tasks in queue have finished
                self.finished.emit(self.downloaded_cnt)
                self.downloaded_cnt = 0

    def resource_task_callback(self, task):
        self.logger.debug('%s from total %d completed with callback'
                          % (task.get_name(), len(self.resource_tasks)))

        # Remove finished task from the queue
        self.resource_tasks.remove(task)

    async def resource_download_async(self, image_url) -> None:
        """
        Downloading resource which will be processed by resource_downloaded_handler() then.
        """
        self.download_image(image_url)

    @staticmethod
    def is_external_url(url, base_domain='example.com'):
        """
        Determine if the given URL is an external URL.
        """
        parsed_url = urlparse(url)
        return not (parsed_url.netloc == '' or parsed_url.netloc.endswith(base_domain))

    @staticmethod
    def url_to_filename(url):
        """
        Generate a filename from a URL by hashing the URL except its basename and combining it with the basename.
        Convert a URL into a safe filename by replacing non-alphanumeric characters with underscores.
        """
        # Extract the basename
        basename = os.path.basename(url)

        # Hash the URL excluding the basename
        url_hash = hashlib.sha256(url.replace(basename, '').encode('utf-8')).hexdigest()
        short_hash = url_hash[:8]  # Use only first 8 characters of the hash

        # Combine the short hash with the basename
        safe_basename = basename.replace(os.path.splitext(basename)[1], '')
        safe_basename = re.sub(r'\W+', '_', safe_basename)  # sanitize to make sure it's a valid file name part
        return f"{short_hash}_{safe_basename}{os.path.splitext(basename)[1]}"

    @staticmethod
    def mime_to_extension(mime_type):
        # Map of common image MIME types to file extensions
        mime_map = {
            'image/jpeg': '.jpg',
            'image/png': '.png',
            'image/gif': '.gif',
            'image/bmp': '.bmp',
            'image/svg+xml': '.svg',
            'image/tiff': '.tiff',
            'image/webp': '.webp'
        }
        return mime_map.get(mime_type, '.dat')  # Default to .dat if MIME type is unknown



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Enum class for the app's supported languages.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import QLocale

from .enum_base import EnumBase

assert hasattr(EnumBase, 'default'), "Check default() method is implemented in the base class"


class Languages(EnumBase):

    DE = "German"
    EN = ("English", True)  # Fallback default option
    ES = "Spanish"
    FI = "Finnish"
    FR = "French"
    GE = "Georgian"
    GR = "Greek"
    IN = "Hindi"
    IT = "Italian"
    JA = "Japanese"
    KO = "Korean"
    LA = "Latin"
    NL = "Dutch"
    PT = "Portuguese"
    RU = "Russian"
    SE = "Swedish"
    TR = "Turkish"
    ZH = "Chinese"

    @classmethod
    def default(cls):
        """
        Returns the default value of the Enum.
        Usage:
            self.default()
        @return: Any
        """

        locale = QLocale.system()  # Get the system's locale
        iso_language_code = locale.name().split('_')[0]  # Extract the language part and ignore the country part

        # Find and return the default enum member based on the system locale
        for member in cls:
            # System locale as a default locale
            if iso_language_code and member.name.lower() == iso_language_code:
                return member

        # Fallback to the parent's logic
        return super().default()



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Provides app search UI form.
- Functionality: Displays the app's search form.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt, QSize, Signal
from PySide6.QtWidgets import QWidget, QLabel, QLineEdit, QPushButton, QHBoxLayout, QSizePolicy
from PySide6.QtGui import QColor

from . import Settings
from . import Lexemes
from . import ThemeHelper

from typing import TYPE_CHECKING, Any, List, Dict, Union  # noqa: F401

import logging

if TYPE_CHECKING:
    from ..ui.toolbar import ToolBar  # noqa: F401


class SearchForm(QWidget):
    """
    Custom search form widget that integrates a QLineEdit for input, QLabel for displaying input labels,
    and other elements to show the actual label and count of search results. This widget also exposes signals
    similar to a QLineEdit for external interaction.
    """

    # Signals to mimic QLineEdit's textChanged and returnPressed events.
    textChanged = Signal(str)
    returnPressed = Signal()
    searchButtonClear = Signal()
    searchButtonNext = Signal()
    searchButtonPrev = Signal()
    caseSensitive = Signal(Qt.CheckState)

    def __init__(self, parent):
        super().__init__()

        self.parent = parent  # type: Union[QWidget, ToolBar]

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the parent widget to the dialog
            self.setFont(self.parent.font())

        self.logger = logging.getLogger('search_form')

        self.settings = Settings()

        self.theme_helper = ThemeHelper()

        self.lexemes = Lexemes(self.settings.app_language, default_scope='toolbar')

        # Placeholder attributes for component widgets, initialized in init_ui.
        self._search_input = None  # type: Union[QLineEdit, None]
        self._search_pos_label = None  # type: Union[QLabel, None]
        self._search_count_label = None  # type: Union[QLabel, None]

        self.case_sensitive_state = False
        # Search navigation button variables will be set via mapping

        self.init_ui()

    def get_toolbar_search_buttons(self) -> List[Dict[str, Any]]:
        """
        Retrieves the configuration map for search-related toolbar buttons.

        Note: The 'var_name' parameter will be initialized in the final object that utilizes this map.

        Returns:
            List[Dict[str, Any]]: A list of dictionaries detailing the configuration of each toolbar search button.
        """

        return [
            {'type': 'action', 'name': 'search_clear', 'system_icon': 'window-close', 'theme_icon': 'x-circle-fill.svg',
             'action': self.searchButtonClear.emit, 'enabled': False, 'default': False,
             'tooltip': self.lexemes.get('search_buttons_label_clear', scope='toolbar'),
             'accessible_name': self.lexemes.get('search_buttons_accessible_name_clear', scope='toolbar'),
             'var_name': 'btn_search_clear', 'color': self.theme_helper.get_color('toolbar_search_button_clear')},
            {'type': 'action', 'name': 'search_prev', 'system_icon': 'go-up', 'theme_icon': 'caret-up-fill.svg',
             'action': self.searchButtonPrev.emit, 'enabled': False, 'default': False,
             'tooltip': self.lexemes.get('search_buttons_label_prev', scope='toolbar'),
             'accessible_name': self.lexemes.get('search_buttons_accessible_name_prev', scope='toolbar'),
             'var_name': 'btn_search_prev', 'color': self.theme_helper.get_color('toolbar_search_button_prev')},
            {'type': 'action', 'name': 'search_next', 'system_icon': 'go-down', 'theme_icon': 'caret-down-fill.svg',
             'action': self.searchButtonNext.emit, 'enabled': False, 'default': True,
             'tooltip': self.lexemes.get('search_buttons_label_next', scope='toolbar'),
             'accessible_name': self.lexemes.get('search_buttons_accessible_name_next', scope='toolbar'),
             'var_name': 'btn_search_next', 'color': self.theme_helper.get_color('toolbar_search_button_next')},
            {'type': 'action', 'name': 'search_case_sensitive', 'system_icon': '', 'theme_icon': None,
             'var_name': 'btn_search_case_sensitive', 'enabled': True, 'default': False, 'color': None,
             'tooltip': self.lexemes.get('search_case_sensitive_tooltip', scope='toolbar'),
             'accessible_name': self.lexemes.get('search_case_sensitive_accessible_name', scope='toolbar'),
             'text': 'aA',  # the 'aA' notation is commonly used universally to depict case sensitivity
             'action': self.case_sensitive_toggle},
        ]

    def init_ui(self):
        # Set up the main layout and internal components
        search_layout = QHBoxLayout(self)
        # Adjust layout margins for proper spacing inside the toolbar
        search_layout.setContentsMargins(0, 0, 0, 0)
        search_layout.setSpacing(0)

        # Searched text
        self._search_input = QLineEdit(self)
        self._search_input.setObjectName('search_input')
        self._search_input.setReadOnly(False)
        self._search_input.setMaxLength(512)
        self._search_input.setPlaceholderText(self.lexemes.get('search_input_placeholder_text'))
        self._search_input.setAccessibleDescription(self.lexemes.get('search_input_accessible_description'))

        # Connect the QLineEdit's textChanged signal to the custom widget's textChanged signal
        self._search_input.textChanged.connect(self.emit_text_changed)
        self._search_input.returnPressed.connect(self.returnPressed.emit)

        # Count occurrences of the searched text within the document.
        self._search_count_label = QLabel('', self)
        self._search_count_label.setObjectName('search_count_label')
        self._search_count_label.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        self._search_count_label.setAlignment(Qt.AlignmentFlag.AlignCenter)  # Center text in QLabel

        # The position of the current occurrence of the searched text within the document.
        self._search_pos_label = QLabel('', self)
        self._search_pos_label.setObjectName('search_pos_label')
        self._search_pos_label.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        self._search_pos_label.setAlignment(Qt.AlignmentFlag.AlignCenter)  # Center text in QLabel

        # Add UI component to the layout
        search_layout.addWidget(self._search_input)

        # Adjust the display properties of labels showing the number of searched text occurrences.
        self.set_counter_text('')

        # Ensure this QLabel is visually layered above other widgets if overlapping.
        self._search_pos_label.raise_()
        self._search_count_label.raise_()

        for search_button in self.get_toolbar_search_buttons():
            if search_button['type'] == 'action':
                # Create and append toolbar button
                self.append_toolbar_button(button_conf=search_button, layout=search_layout)

        # Connect to the event that clears the search form
        self.searchButtonClear.connect(self.action_search_clear)

    def action_search_clear(self) -> None:
        """
        Clears the search field, resets case sensitivity, and updates the interface
        to reflect the cleared state.
        """

        # Clears the search field by setting the current search text to an empty string.
        self.set_text('')

        # Reset case sensitivity.
        # Unused: This line is commented out to allow independent handling of the button's toggle state.
        # self.set_case_sensitive(False)

        # Refresh the state of search buttons
        self.action_search_buttons_refresh()

        # Hide the occurrences counter
        self.set_counter_text('')

    def action_search_buttons_refresh(self):
        """
        Updates the enable/disable state of the search buttons based on the presence of search text and occurrences.
        """

        try:
            search_occurrences = int(self.counter_text())
        except ValueError:
            search_occurrences = 0

        # Retrieve the current search text.
        text = self.text()
        if hasattr(self, 'btn_search_clear'):
            self.btn_search_clear.setEnabled(True if len(text) > 0 else False)

        # Determine the state for navigation buttons based on occurrences.
        btn_state = bool(search_occurrences)
        if hasattr(self, 'btn_search_next'):
            self.btn_search_next.setEnabled(btn_state)
        if hasattr(self, 'btn_search_prev'):
            self.btn_search_prev.setEnabled(btn_state)

        # Highlight the case-sensitive button if that mode is active.
        if hasattr(self, 'btn_search_case_sensitive'):
            self.search_button_highlight(button=self.btn_search_case_sensitive, state=self.case_sensitive())

    def case_sensitive_toggle(self):
        """
        This method switches the state of case sensitivity and notifies other components
        of this change by emitting a caseSensitive signal with the current state.
        """

        # Toggle the case sensitivity state
        self.case_sensitive_state ^= True

        # Emit the caseSensitive signal with the current state to notify other components
        self.caseSensitive.emit(self.case_sensitive_state)

    def search_button_highlight(self, button: QPushButton, state: bool = False):
        if state:
            button.setStyleSheet("QPushButton { background: %s; }"
                                 % self.theme_helper.get_color('toolbar_search_button_highlight', True))
        else:
            button.setStyleSheet("QPushButton { }")

    def append_toolbar_button(self, button_conf: dict, layout: QHBoxLayout) -> None:
        """
        Appends a search button to the toolbar based on the provided configuration.

        Args:
            button_conf (dict): Configuration dictionary for the button, specifying properties like text, icons, tooltips, etc.
            layout (QHBoxLayout): The layout to which the button will be added.

        This helper function creates and adds a search button to the specified layout of the toolbar.
        """

        system_icon = button_conf['system_icon'] if 'system_icon' in button_conf else None
        theme_icon = button_conf['theme_icon'] if 'theme_icon' in button_conf else None
        theme_icon_color = QColor(button_conf['color']) if 'color' in button_conf \
            else self.theme_helper.get_color('toolbar_icon_color_default')

        icon = None
        if theme_icon:
            # Get the theme icon, using a system default as a fallback.
            icon = self.theme_helper.get_icon(theme_icon=theme_icon, system_icon=system_icon, color=theme_icon_color)

        # Initialize a toolbar button with an icon.
        icon_button = QPushButton(self)
        if 'name' in button_conf:
            icon_button.setObjectName(button_conf['name'])  # Set object name from configuration.
        icon_button.setFont(self.font())

        # Match button size to the search input field's height, maintaining aspect ratio.
        size_hint = self._search_input.sizeHint()
        icon_button.setMinimumSize(QSize(size_hint.height(), size_hint.height()))

        # Configure button to display either an icon or text based on configuration.
        if icon:
            # Calculate and set icon size based on button height.
            icon_width = icon_height = int(size_hint.height() * 0.7)
            icon_button.setIconSize(QSize(icon_width, icon_height))
            icon_button.setIcon(icon)  # Set the icon.
        elif 'text' in button_conf and button_conf['text']:
            # Adjust and set the button's font size.
            icon_button_font = icon_button.font()
            icon_button_font.setPointSizeF(icon_button_font.pointSize() * 0.8)
            icon_button.setFont(icon_button_font)
            icon_button.setText(button_conf['text'])  # Set button text.

        action = button_conf['action'] if 'action' in button_conf else None  # Action when clicked
        if action is not None and callable(action):
            icon_button.clicked.connect(action)
        if 'accessible_name' in button_conf:
            icon_button.setAccessibleName(button_conf['accessible_name'])
        if 'tooltip' in button_conf:
            icon_button.setToolTip(button_conf['tooltip'])
        if 'default' in button_conf:
            icon_button.setDefault(button_conf['default'])
        if 'enabled' in button_conf:
            icon_button.setEnabled(button_conf['enabled'])

        # Add the button to the layout
        layout.addWidget(icon_button)

        if 'var_name' in button_conf:
            if hasattr(self, button_conf['var_name']):
                self.logger.debug('Variable "%s" is already set! Re-writing it...' % button_conf['var_name'])
            setattr(self, button_conf['var_name'], icon_button)  # type: QPushButton

    def emit_text_changed(self, text):
        """
        Emit the custom textChanged signal with the current text as the argument to notify of text changes.
        """
        self.textChanged.emit(text)

    def text(self):
        """
        Retrieve the current text from the underlying QLineEdit.

        Returns:
            str: The text currently displayed in the QLineEdit.
        """
        return self._search_input.text()

    def set_text(self, text):
        """
        Proxy method to set the text of the underlying QLineEdit to the specified value.

        Args:
            text (str): The text to display in the QLineEdit.

        Returns:
            None: This method does not return a value.
        """
        return self._search_input.setText(text)

    def case_sensitive(self) -> bool:
        """
        Determine whether the case-sensitive search option is enabled in the search settings.

        Returns:
            bool: True if the case-sensitive search is enabled, False otherwise.
        """

        return self.case_sensitive_state

    def set_case_sensitive(self, checked: bool = False):
        """
        Enable or disable the case-sensitive search option based on the checked parameter.

        Args:
            checked (bool): If True, enables case-sensitive search; if False, disables it.

        Returns:
            None: This method does not return a value.
        """

        self.case_sensitive_state = True if checked else False

    def set_focus(self):
        self._search_input.setFocus()

    def set_maximum_width(self, width):
        self.setMaximumWidth(width)

    def set_placeholder_text(self, text):
        self._search_input.setPlaceholderText(text)

    def counter_text(self):
        """
        Retrieve the text from the label displaying the count of searched text occurrences.

        Returns:
            str: The text indicating the number of occurrences of the searched text.
        """
        return self._search_count_label.text()

    def set_counter_text(self, count: str):
        """
        Sets the text of the embedded QLabel which displays the count of searched text occurrences.

        Args:
            count (str): The text to display in the count label; representing the number of occurrences.
        """

        self._search_count_label.setText(count)
        self.adjust_occurrence_label_size(label=self._search_count_label)
        self.set_position_text('')
        # Refresh the state of search buttons based on current search conditions.
        self.action_search_buttons_refresh()

    def position_text(self):
        """
        Retrieve the text from the label displaying the current position of the searched text occurrence.

        Returns:
            str: The text indicating the current position of occurrences of the searched text.
        """
        return self._search_pos_label.text()

    def set_position_text(self, index: str):
        """
        Sets the text of the embedded QLabel that shows the position of the current search result.

        Args:
            index (str): The position of the current search result to display.
                         If this is '0', the label will be hidden.
        """

        try:
            pos_text = index if len(index) > 0 and int(index) > 0 else ''
        except ValueError:
            pos_text = ''

        self._search_pos_label.setText(pos_text)
        self.adjust_occurrence_label_size(label=self._search_pos_label, shift=self._search_count_label.width())

    def adjust_occurrence_label_size(self, label: QLabel, shift: int = 0):
        """
        Adjusts the visibility and dimensions of a QLabel based on its content and related UI elements.

        The label's visibility is controlled by its content (visible if not empty).
        Its size and position are dynamically adjusted to ensure proper layout within the interface,
        particularly in relation to the search input field.

        Args:
            label (QLabel): The label whose size and visibility are to be adjusted.
            shift (int, optional): The horizontal offset to apply when positioning the label. Defaults to 0.
        """

        # Show occurrence label if not an empty text
        try:
            visible = True if len(label.text()) > 0 and int(label.text()) >= 0 else False
        except ValueError:
            visible = False

        label.setVisible(visible)

        if not visible:
            return

        # First, calculate the width of the count label
        digits = max(1, len(str(label.text())))
        # '10' represents the total left and right padding
        space = 10 + self.fontMetrics().horizontalAdvance("0") * digits
        # Set the size of QLabel to achieve the desired overlap
        label.resize(space, self._search_input.height())

        # Adjust the label's size and position dynamically when the widget is resized
        label.move(self._search_input.width() - label.width() - shift, self._search_input.y())
        self._search_pos_label.resize(label.width(), self._search_input.height())

    def resizeEvent(self, event):
        super().resizeEvent(event)
        # Adjust the size of the count label
        self.adjust_occurrence_label_size(label=self._search_count_label)
        self.adjust_occurrence_label_size(label=self._search_pos_label, shift=self._search_count_label.width())



# Italian lexemes ai_assistant.py
lexemes = {
    # Assistente IA
    "dialog_title": "Richiesta all'IA",
    "dialog_prompt_input_placeholder_text": "Richiesta",
    "dialog_prompt_input_accessible_description": "Inserisci richiesta per fornire contesto all'assistente IA",
    "dialog_response_output_placeholder_text": "Dati di risposta",
    "dialog_response_output_accessible_description": "I dati di risposta appariranno in questo campo",
    "dialog_response_output_notice_empty_text": "Fornisci un input di testo per ricevere una risposta di circa...",

    "dialog_usage_module_label": "MODULO",
    "dialog_usage_model_label": "MODELLO",
    "dialog_usage_tokens_label": "TOKEN",
    "dialog_usage_tokens_prompt": "prompt: {tokens}",
    "dialog_usage_tokens_answer": "risposta: {tokens}",
    "dialog_usage_tokens_total": "totale: {tokens}",

    "dialog_button_send_request": "Invia richiesta",
    "dialog_button_stop_request": "Stop",
    "dialog_button_save_history": "Salva la cronologia della conversazione",

    "dialog_message_copy_tooltip": "Copia testo",
    "dialog_message_copied_tooltip": "Copiato",

    "dialog_error_loading_model": "Errore nel caricamento del modello personalizzato",
}



# Latin lexemes toolbar.py
lexemes = {
    "actions_label_new_document": "Documentum novum",
    "actions_accessible_name_new_document": "Documentum novum",
    "actions_label_open": "Aperire",
    "actions_accessible_name_open": "Documentum aperire",
    "actions_label_save": "Salvare",
    "actions_accessible_name_save": "Documentum salvare",
    "actions_label_save_as": "Salvare ut",
    "actions_accessible_name_save_as": "Salvare ut exemplar documenti",
    "actions_label_sync": "Synchronizare",
    "actions_accessible_name_sync": "Documentum cum reposito externo synchronizare",
    "actions_label_edit": "Recensere",
    "actions_accessible_name_edit": "Fasciculum recensere",
    "actions_label_source": "Fons",
    "actions_accessible_name_source": "Codex",
    "actions_label_encrypt": "Codicare",
    "actions_accessible_name_encrypt": "Fasciculum codicare",
    "actions_label_decrypt": "Decodicare",
    "actions_accessible_name_decrypt": "Fasciculum decodicare",
    "actions_label_bold": "Audax",
    "actions_accessible_name_bold": "Formatum textus audax",
    "actions_label_italic": "Italicus",
    "actions_accessible_name_italic": "Formatum textus italicus",
    "actions_label_underline": "Sublineare",
    "actions_accessible_name_underline": "Formatum textus sublineare",
    "actions_label_strikethrough": "Exstinguere",
    "actions_accessible_name_strikethrough": "Formatum textus exstinguere",
    "actions_label_blockquote": "Blockquote",
    "actions_accessible_name_blockquote": "Formatum textus blockquote",
    "actions_label_color_picker": "Color selector",
    "actions_accessible_name_color_picker": "Elige colorem textus",
    "actions_label_ai_assistant": "Assistens AI",
    "actions_accessible_name_ai_assistant": "Assistentem AI exercere",
    "actions_label_about": "De",
    "actions_accessible_name_about": "De App",
    "actions_label_exit": "Exire",
    "actions_accessible_name_exit": "Exire App",
    "actions_label_settings": "Configurationes",
    "actions_accessible_name_settings": "Configurationes Applicatio",

    "search_input_label": "Quaerere",
    "search_input_placeholder_text": "Quaere textum in documento",
    "search_input_accessible_description": "Campus input quaerere",
    "search_buttons_label_clear": "Quaestionem purare",
    "search_buttons_accessible_name_clear": "Campum quaestionis purare",
    "search_buttons_label_prev": "Quaestio praecedens",
    "search_buttons_accessible_name_prev": "Eventus quaestionis praecedentis",
    "search_buttons_label_next": "Quaestio proxima",
    "search_buttons_accessible_name_next": "Eventus quaestionis proximae",
    "search_case_sensitive_accessible_name": "Casu sensitivo",
    "search_case_sensitive_tooltip": "Quaerere casu sensitivo",
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Extends QTextBlockUserData to support data storage for text editing cross-actions.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtGui import QTextBlockUserData

from typing import Any, Type


class TextBlockData(QTextBlockUserData):
    def __init__(self, block_number):
        super().__init__()

        self.block_number = block_number
        self.data = {}

    def put(self, tag=str, opened=bool, within=bool, closed=bool, start: int = 0, end: int = 0) -> None:
        new_data = {'opened': opened, 'within': within, 'closed': closed, 'start': start, 'end': end}
        existing_data = self.get_all(tag)
        if existing_data is not None:
            for data_index, data_row in enumerate(existing_data):
                # The equal element is already exist
                if data_row == new_data:
                    return
                # Check either update needed or not
                # TODO: Improve the check algorithm as it's quite simple at the moment.
                if data_row['start'] == new_data['start'] and data_row['end'] == new_data['end']:
                    existing_data[data_index].update(new_data)
            if isinstance(existing_data, list):
                existing_data.append(new_data)
            elif isinstance(existing_data, dict):
                existing_data = [existing_data, new_data]
            else:
                existing_data = [new_data]
            self.data[tag] = existing_data
        else:
            self.data[tag] = [new_data]

    def update(self, tag=str, index=Type[int], opened=bool, within=bool, closed=bool,
               start: int = 0, end: int = 0) -> None:
        existing_data = self.get_one(tag, index)
        if existing_data is not None:
            existing_data.update({'opened': opened, 'within': within, 'closed': closed, 'start': start, 'end': end})
            self.data[tag][index] = existing_data

    def get_one(self, tag=str, index=0) -> Any:
        """
        This method returns a single piece of list data.
        Consider using it with code blocks only where the multiple occurrences are not supported.
        """
        if (tag in self.data
                and self.data[tag]
                and 0 <= index < len(self.data[tag])):
            return self.data[tag][index]
        return None

    def get_all(self, tag=str) -> Any:
        """
        This method returns all stored pieces of list data.
        Consider using it with the multiple occurrences, like a few strikethrough elements within one line.
        """
        if tag in self.data and self.data[tag]:
            return self.data[tag]
        return None

    def get_param(self, tag=str, param=str) -> Any:
        item = self.get_one(tag)
        if item and param in item:
            return item[param]
        return None

    def drop(self, tag=str):
        if tag in self.data:
            self.data.__delitem__(tag)

    def drop_index(self, tag=str, index=0) -> None:
        if tag in self.data:
            if 0 <= index < len(self.data[tag]):
                # These variants are also work:
                # self.data[tag].__delitem__(index)
                # self.data[tag] = [item for key, item in enumerate(self.data[tag]) if key != index]
                del self.data[tag][index]

    def search(self, tag=str, key=str, value=any) -> Any:
        if tag in self.data:
            return [data for data in self.data[tag] if key in data and data[key] == value][0]
        return None

    def __repr__(self):
        return '\n'.join([f'[{self.block_number}] {tag}: {item}' for tag, item in self.data.items()])



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: AI Assistant Dialog Class.
- Functionality: This class facilitates interactions between the user and various external plugins or APIs.
  It acts as a bridge, sending queries to and receiving responses from selected plugins, ensuring seamless integration
  and efficient data handling. It dynamically leverages user inputs and contextual information for API interactions.

Features:
- Sends requests and receives responses based on user inputs or contextual cues.
- Implements robust error handling to manage API limitations or failures, ensuring consistent application performance.
- Adaptable functionality to suit different application needs.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt, QSize, Signal, Slot
from PySide6.QtWidgets import QDialog, QVBoxLayout, QWidget, QPushButton
from PySide6.QtWidgets import QLabel, QSizePolicy, QHBoxLayout, QScrollArea
from PySide6.QtGui import QPixmap, QColor

import logging

from typing import TYPE_CHECKING, Union

from . import Settings
from . import Lexemes
from . import ThemeHelper

from .ai_message_label import AIMessageLabel
from .ai_prompt_input import AIPromptInput

from ..vertical_line_spacer import VerticalLineSpacer
from ..rotating_label import RotatingLabel

from ...modules.modules import Modules
from ...modules.base_ai_core import BaseAiCore
from ...enums.enum_base import EnumBase

from qasync import asyncClose
from datetime import datetime

import markdown

if TYPE_CHECKING:
    from PySide6.QtWidgets import QScrollBar  # noqa: F401


class EnumMessageType(EnumBase):
    # Sync with corresponding CSS-styles
    DEFAULT = ("default", True)  # Fallback default option
    USER_INPUT = "user_input"
    RESPONSE = "response"


class EnumMessageStyle(EnumBase):
    DEFAULT = ("default", True)  # Fallback default option
    INFO = "info"
    ERROR = "error"


class AIAssistant(QDialog):

    dialog_closed = Signal()
    request_cancelled = Signal()

    message_added = Signal(str, int, int, EnumMessageType)

    md: markdown.Markdown = None

    module: None
    # Inference module core
    module_core: BaseAiCore = None

    # Set the inference status
    is_in_progress = False

    def __init__(self, parent):
        super().__init__(parent, Qt.WindowType.Window)

        # self.setWindowFlags(Qt.WindowType.Window | Qt.WindowType.WindowStaysOnTopHint)
        self.setWindowState(Qt.WindowState.WindowActive)

        # Make the dialog non-modal
        self.setModal(False)  # or self.setWindowModality(Qt.WindowModality.NonModal)

        # Parent instance
        self.parent = parent

        # Apply font from the dialog instance to the label
        self.setFont(self.parent.font())

        self.logger = logging.getLogger('ai_assistant')

        self.settings = Settings(parent=self)
        self.settings.value_changed.connect(
            lambda v: self.settings_update_handler(v))

        # Load modules first to enable the loading of extension settings.
        module_instances = []
        for module in Modules().get_by_extension('ai_assistant'):
            # Pass settings object to avoid circular dependencies
            module_instances.append(Modules().create(module))

        # A module to use for inference
        self.inference_module = self.settings.ai_config_inference_module
        try:
            # Get the name of the inference module
            self.inference_module_name = Modules().modules.get(self.inference_module).get_name()
        except Exception as e:
            self.logger.warning(f"Inference module '{self.inference_module}' name is not set {e}")
            # Try to find any suitable module
            for module_name in self.settings.ai_config_inference_modules.keys():
                # Set up first available module as a default one
                self.settings.ai_config_inference_module = module_name
                self.inference_module = self.settings.ai_config_inference_module
                break
            # Update with the default module name
            self.inference_module_name = self.inference_module
            self.logger.warning(f"Fallback to the inference module named '{self.inference_module}'")

        self.theme_helper = ThemeHelper()

        # Load lexemes for selected language and scope
        self.lexemes = Lexemes(self.settings.app_language, default_scope='ai_assistant')

        # Dialog layout
        self.layout = QVBoxLayout(self)

        self.setWindowTitle(self.lexemes.get('dialog_title'))

        # UI element variables
        self.prompt_input = None  # type: Union[AIPromptInput, None]
        self.messages_area = None  # type: Union[QScrollArea, None]
        self.messages_scroll = None  # type: Union[QScrollBar, None]
        self.messages_layout = None  # type: Union[QVBoxLayout, None]
        self.background_label = None  # type: Union[QLabel, None]
        self.module_name_label = None  # type: Union[QLabel, None]
        self.model_name_label = None  # type: Union[QLabel, None]
        self.tokens_prompt_label = None  # type: Union[QLabel, None]
        self.tokens_answer_label = None  # type: Union[QLabel, None]
        self.tokens_total_label = None  # type: Union[QLabel, None]
        self.send_button = None  # type: Union[QPushButton, None]
        self.save_history_button = None  # type: Union[QPushButton, None]

        # Last added message id
        self.message_id = 0

        # Last requested message id
        self.request_message_id = 0

        # Message ids
        self.message_ids = []

        # Prompt and response tokens
        self.token_usage = {}

        self.init_ui()

    def init_ui(self):
        # Set dialog size derived from the main window size
        main_window_size = self.parent.size()
        dialog_width = int(main_window_size.width() * 0.7)
        dialog_height = int(main_window_size.height() * 0.7)
        # self.setMinimumSize(dialog_width, dialog_height)
        self.resize(dialog_width, dialog_height)

        # Main view margins
        self.setContentsMargins(5, 5, 5, 5)
        # Corresponding styles
        self.setStyleSheet(self.theme_helper.get_css('ai_assistant'))

        self.messages_area = QScrollArea(self)
        self.messages_area.setWidgetResizable(True)
        self.messages_area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.messages_area.setStyleSheet("QScrollArea { border: none; }")
        self.messages_area.setAccessibleDescription(
            self.lexemes.get('dialog_response_output_accessible_description'))

        # Initialize the vertical scrollbar for the messages area
        self.messages_scroll = self.messages_area.verticalScrollBar()
        # Connect the rangeChanged signal to ensure the messages area scrolls to the bottom
        self.messages_scroll.rangeChanged.connect(self.scroll_to_bottom)

        scroll_widget = QWidget()
        self.messages_layout = QVBoxLayout(scroll_widget)
        self.messages_layout.setContentsMargins(0, 0, 0, 5)
        self.messages_layout.setSpacing(0)
        # scroll_widget.setLayout(self.messages_layout)
        self.messages_area.setWidget(scroll_widget)

        self.layout.addWidget(self.messages_area)

        # To fill the space atop
        spacer_widget = QWidget()
        spacer_widget.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.messages_layout.addWidget(spacer_widget)

        # Prompt input field
        self.prompt_input = AIPromptInput(send_callback=self.send_request, parent=self)
        self.layout.addWidget(self.prompt_input)

        label_size = QSize(48, 48)
        label_pixmap = QPixmap(self.theme_helper.get_icon(
            theme_icon='clock-fill.svg',
            color=QColor(self.theme_helper.get_color('ai_assistant_status_icon_color', css_format=True)))
                               .pixmap(label_size))
        # Use custom Pixmap class to assure the transformation (rotation)
        self.background_label = RotatingLabel(pixmap=label_pixmap, parent=self)
        self.background_label.resize(label_size)
        # self.background_label.setPixmap(label_pixmap)
        self.background_label.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
        # Set at the center of the result text field
        self.background_label.setGeometry(self.width() // 2 - self.background_label.width() // 2,
                                          self.height() // 2 - self.background_label.height() // 2,
                                          self.background_label.width(),
                                          self.background_label.height())
        # Rise it up on the first plane
        self.background_label.raise_()
        # Hide on the very beginning as it will be set to visible upon request
        self.background_label.hide()

        status_bar_widget = QWidget(self)
        status_bar_layout = QHBoxLayout(status_bar_widget)

        model_label = QLabel(text=self.lexemes.get('dialog_usage_module_label'))
        model_label.setObjectName("ai_assistant_dialog_usage_module_name_label")
        model_label.setFont(self.font())
        model_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        model_label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        # model_label.setStyleSheet("QLabel {color: grey;}")
        status_bar_layout.addWidget(model_label)

        status_bar_layout.addWidget(VerticalLineSpacer())

        self.module_name_label = QLabel(self.inference_module_name)
        self.module_name_label.setFont(self.font())
        self.module_name_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.module_name_label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        self.module_name_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        status_bar_layout.addWidget(self.module_name_label)

        self.model_name_label = QLabel()  # Text will be updated later
        self.model_name_label.setObjectName("ai_assistant_dialog_usage_model_name_label")
        self.model_name_label.setFont(self.font())
        self.model_name_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        self.model_name_label.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        self.model_name_label.setAlignment(Qt.AlignmentFlag.AlignLeft)
        status_bar_layout.addWidget(self.model_name_label)

        central_spacer = QWidget()
        central_spacer.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)
        status_bar_layout.addWidget(central_spacer)

        tokens_label = QLabel(text=self.lexemes.get('dialog_usage_tokens_label'))
        tokens_label.setFont(self.font())
        tokens_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        tokens_label.setObjectName("ai_assistant_dialog_usage_tokens_label")
        # tokens_label.setStyleSheet("QLabel {color: grey;}")
        status_bar_layout.addWidget(tokens_label)
        status_bar_layout.addWidget(VerticalLineSpacer(), alignment=Qt.AlignmentFlag.AlignRight)
        self.tokens_prompt_label = QLabel()
        self.tokens_prompt_label.setFont(self.font())
        self.tokens_prompt_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        status_bar_layout.addWidget(self.tokens_prompt_label)
        status_bar_layout.addWidget(VerticalLineSpacer(), alignment=Qt.AlignmentFlag.AlignRight)
        self.tokens_answer_label = QLabel()
        self.tokens_answer_label.setFont(self.font())
        self.tokens_answer_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        status_bar_layout.addWidget(self.tokens_answer_label)
        status_bar_layout.addWidget(VerticalLineSpacer(), alignment=Qt.AlignmentFlag.AlignRight)
        self.tokens_total_label = QLabel()
        self.tokens_total_label.setFont(self.font())
        self.tokens_total_label.setAlignment(Qt.AlignmentFlag.AlignRight)
        status_bar_layout.addWidget(self.tokens_total_label)

        self.layout.addWidget(status_bar_widget)
        # Update usage with initial params
        self.update_usage('')

        buttons_layout = QHBoxLayout()
        buttons_widget = QWidget()
        buttons_widget.setLayout(buttons_layout)

        # Save button
        self.save_history_button = QPushButton()
        self.save_history_button.setFont(self.font())
        self.save_history_button.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.save_history_button.setIcon(self.theme_helper.get_icon(
            theme_icon='floppy2.svg',
            color=QColor(self.theme_helper.get_color('ai_assistant_button_icon_color', css_format=True))))
        self.save_history_button.clicked.connect(self.save_history_action)
        self.save_history_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.save_history_button.setDisabled(True)
        self.save_history_button.setToolTip(self.lexemes.get('dialog_button_save_history'))
        buttons_layout.addWidget(self.save_history_button)

        # To fill the space between
        spacer_widget = QWidget()
        spacer_widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        buttons_layout.addWidget(spacer_widget, 3)

        # Submit button
        self.send_button = QPushButton(self.lexemes.get('dialog_button_send_request'), self)
        self.send_button.setFont(self.font())
        self.send_button.setIcon(self.theme_helper.get_icon(
            theme_icon='play-fill.svg',
            color=QColor(self.theme_helper.get_color('ai_assistant_button_icon_color', css_format=True))))
        self.send_button.setCursor(Qt.CursorShape.PointingHandCursor)
        self.send_button.clicked.connect(self.send_request)
        buttons_layout.addWidget(self.send_button, 1)

        self.layout.addWidget(buttons_widget)

        # Set the main layout for the dialog
        self.setLayout(self.layout)

    def get_message_color(self, message_type: EnumMessageType, message_style: EnumMessageStyle) -> tuple:
        if message_style == EnumMessageStyle.INFO:
            color = self.theme_helper.get_color('ai_assistant_message_info_color', True)
            bg_color = self.theme_helper.get_color('ai_assistant_message_info_background_color', True)
        elif message_style == EnumMessageStyle.ERROR:
            color = self.theme_helper.get_color('ai_assistant_message_error_color', True)
            bg_color = self.theme_helper.get_color('ai_assistant_message_error_background_color', True)
        else:
            if message_type == EnumMessageType.USER_INPUT:
                color = self.theme_helper.get_color('ai_assistant_message_user_input_color', True)
                bg_color = self.theme_helper.get_color('ai_assistant_message_user_input_background_color', True)
            elif message_type == EnumMessageType.RESPONSE:
                color = self.theme_helper.get_color('ai_assistant_message_response_color', True)
                bg_color = self.theme_helper.get_color('ai_assistant_message_response_background_color', True)
            else:
                color = self.theme_helper.get_color('ai_assistant_message_default_color', True)
                bg_color = self.theme_helper.get_color('ai_assistant_message_default_background_color', True)
        return color, bg_color

    def add_message(self, message, request_msg_id=None, response_msg_id=None,
                    message_type: EnumMessageType = EnumMessageType.default(),
                    message_style: EnumMessageStyle = EnumMessageStyle.default()) -> int:
        # Message background color
        color, bg_color = self.get_message_color(message_type, message_style)

        if message_type == EnumMessageType.USER_INPUT:
            if request_msg_id is None:
                # Message id might not be set in case of a new message needed
                request_msg_id = self.gen_next_message_id()
            message_id = request_msg_id
        elif message_type == EnumMessageType.RESPONSE:
            if response_msg_id is None:
                # Message id might not be set in case of a new message needed
                response_msg_id = self.gen_next_message_id()
            message_id = response_msg_id
        else:
            # Default or info; new internal id
            message_id = self.gen_next_message_id()

        # Remove leading spaces if exist
        message_text = message.lstrip()

        # Create a stylized message label
        message_label = AIMessageLabel(text=message_text, color=color, bg_color=bg_color, parent=self)
        message_label.setObjectName(f'msg_{message_type}_{message_id}')

        self.messages_layout.setAlignment(message_label, Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignBottom)
        self.messages_layout.addWidget(message_label, alignment=Qt.AlignmentFlag.AlignBottom)

        # Emit message added signal
        self.message_added.emit(message_text, request_msg_id, response_msg_id, message_type)

        # Unique message id
        return message_id

    def append_to_message(self, additional_text, request_msg_id=None, response_msg_id=None,
                          message_type: EnumMessageType = EnumMessageType.default(),
                          message_style: EnumMessageStyle = EnumMessageStyle.DEFAULT):
        last_message = None
        if response_msg_id:
            # Message with the same id
            last_message = self.findChild(QLabel, f'msg_{message_type}_{response_msg_id}')
        # Double check found object type
        if isinstance(last_message, QLabel):
            current_text = last_message.text()
            new_text = current_text + additional_text
            last_message.setText(new_text)
            last_message.adjustSize()
        else:
            self.add_message(additional_text, request_msg_id, response_msg_id, message_type, message_style)

        self.messages_area.verticalScrollBar().setValue(self.messages_area.verticalScrollBar().maximum())

    def gen_next_message_id(self) -> int:
        # Update with gaps to allow extra space in case of race condition
        self.message_id += 3
        self.message_ids.append(self.message_id)
        return self.message_id

    def get_prev_message_id(self) -> Union[int, None]:
        # Check if there are at least two items in the list
        if len(self.message_ids) > 1:
            # Return the second-to-last item
            return self.message_ids[-2]
        else:
            # Return None if there are not enough items
            return None

    def init_md(self) -> None:
        """
        Init Markdown object and set it to variable.
        """
        extensions = ['markdown.extensions.extra']
        # Init markdown object with the selected extensions
        self.md = markdown.Markdown(extensions=extensions)

    def convert_markdown_to_html(self, md_content: str) -> str:
        """
        Process Markdown syntax and convert it to html.
        """
        if not self.md:
            self.init_md()
        # Convert markdown to html
        html_content = self.md.convert(md_content)
        # Converted html data
        return html_content

    def update_usage(self, model, prompt_tokens=None, response_tokens=None, total_tokens=None, append=False):
        if not append:
            # Count token usage
            self.token_usage = {model: {'prompt_tokens': 0, 'response_tokens': 0, 'total_tokens': 0}}
        # Update model from response
        if hasattr(self, 'model_name_label'):
            self.model_name_label.setText(model if len(model) > 0 else '')
        # Tokens spent in user's prompt
        prompt_tokens_cnt = prompt_tokens if prompt_tokens else 0
        if append:
            prompt_tokens_cnt += self.token_usage[model]['prompt_tokens']
        self.token_usage[model]['prompt_tokens'] = prompt_tokens_cnt
        if hasattr(self, 'tokens_prompt_label'):
            self.tokens_prompt_label.setText(self.lexemes.get('dialog_usage_tokens_prompt',
                                                              tokens=prompt_tokens_cnt))
        # Tokens spent in response
        response_tokens_cnt = response_tokens if response_tokens else 0
        if append:
            response_tokens_cnt += self.token_usage[model]['response_tokens']
        self.token_usage[model]['response_tokens'] = response_tokens_cnt
        if hasattr(self, 'tokens_answer_label'):
            self.tokens_answer_label.setText(self.lexemes.get('dialog_usage_tokens_answer',
                                                              tokens=response_tokens_cnt))
        # Total tokens spent (combined)
        total_tokens_cnt = total_tokens if total_tokens else 0
        if append:
            total_tokens_cnt += self.token_usage[model]['total_tokens']
        self.token_usage[model]['total_tokens'] = total_tokens_cnt
        if hasattr(self, 'tokens_total_label'):
            self.tokens_total_label.setText(self.lexemes.get('dialog_usage_tokens_total',
                                                             tokens=total_tokens_cnt))

    def init_module(self) -> Union[BaseAiCore, None]:
        self.module = Modules().import_module(self.inference_module)
        if not self.module:
            return None
        if self.module_core:
            return self.module_core
        try:
            # Create an instance of the inference module
            self.module_core = Modules().create(self.module)
            # Connect dialog first before posting any messages to allow correct stream processing
            self.module_core.attach_dialog(self)  # type: ignore
            # Return module core instance
            return self.module_core
        except (RuntimeError, Exception) as e:
            self.logger.warning(f'Error occurred during module {self.inference_module} init {e}')
        return None

    def send_request(self) -> None:
        # If inference is in progress, skip the request
        if self.is_in_progress:
            self.logger.info("Inference in progress. Request skipped.")
            return

        # Check previous request id
        if self.request_message_id and self.request_message_id == self.message_id:
            self.logger.warning(
                    f"Duplicate request message id {self.request_message_id}")
            return

        # Get context for the prompt
        prompt_context = self.prompt_input.toPlainText()
        if len(prompt_context) == 0:
            self.add_message(self.lexemes.get('dialog_response_output_notice_empty_text'), None, None,
                             EnumMessageType.DEFAULT, EnumMessageStyle.INFO)
            return
        # Do not show placeholder text after first input for user convenience
        self.prompt_input.setPlaceholderText('')

        try:
            # Get an instance of the inference module
            module_core = self.init_module()
            # Get a prompt manager
            prompt_manager = module_core.get_prompt_manager()
            # Add request message first to allow it to appear within prompt history
            self.request_message_id = self.add_message(prompt_context, None, None,
                                                       EnumMessageType.USER_INPUT)
            # Either a multi-turn or stateless prompt
            user_prompt = prompt_manager.get_prompt(multi_turn=self.settings.ai_config_multi_turn_dialogue)
            self.logger.debug(f'The prompt:\n{user_prompt}\n')
            self.set_status_waiting()  # Pending status
            # Obtain new message id
            response_msg_id = self.gen_next_message_id()
            # Assume the previous message was a request
            check_request_msg_id = self.get_prev_message_id()
            if self.request_message_id != check_request_msg_id:
                self.logger.warning(
                        f"Request message id doesn't match: {self.request_message_id} vs {check_request_msg_id}")
            module_core.request(
                user_prompt=user_prompt,
                request_msg_id=self.request_message_id,
                response_msg_id=response_msg_id,
                init_callback=self.set_status_processing,  # Processing status
                finished_callback=self.send_request_finished_callback)  # Finished status
        except (AttributeError, ValueError) as e:
            self.logger.warning(f'Error occurred during the request: {e}')
            # Set up response status visible within response field
            self.add_message(self.lexemes.get('dialog_error_loading_model'), None, None,
                             EnumMessageType.DEFAULT, EnumMessageStyle.ERROR)
            self.cancel_request()
            self.set_status_ready()
        except (RuntimeError, Exception) as e:
            self.logger.error(f'Error occurred during the request: {e}')
            self.cancel_request()
            self.set_status_ready()

    def send_request_finished_callback(self, request_msg_id=None, response_msg_id=None,
                                       message_type=EnumMessageType.RESPONSE):
        # Objects might be deleted
        try:
            # Return ready status
            self.set_status_ready()
            # Find the target message
            finished_msg = None
            if response_msg_id:
                # Message (response) with the same id
                finished_msg = self.findChild(QLabel, f'msg_{message_type}_{response_msg_id}')
            # Double check found object type
            if isinstance(finished_msg, QLabel):
                plain_text = finished_msg.text()
                # Check if postprocessing is needed
                if self.settings.ai_config_convert_to_md:
                    # Convert markdown response
                    md_text = self.convert_markdown_to_html(plain_text)
                    finished_msg.setText(md_text)
                # Emit message added signal to update final variant of the message (might not be completed yet)
                self.message_added.emit(plain_text, request_msg_id, response_msg_id, EnumMessageType.RESPONSE)
        except RuntimeError as e:  # Object is already deleted
            self.logger.warning(f'Finished event callback interrupted; probably an object was already deleted: {e}')

    def set_status_waiting(self):
        # Set the inference status
        self.is_in_progress = True
        # Set loader cursor whilst loading content
        self.setCursor(Qt.CursorShape.WaitCursor)
        # Disable prompt input field
        self.prompt_input.setDisabled(True)
        # Clear prompt field as it have to be in the chat already
        self.prompt_input.clear()
        # Disable send request button
        self.send_button.setDisabled(True)
        # Disable save button
        self.save_history_button.setDisabled(True)
        # Show progress label
        self.background_label.show()

    def set_status_processing(self):
        try:
            # Set the inference status
            self.is_in_progress = True
            # Set cursor back
            self.setCursor(Qt.CursorShape.ArrowCursor)
            # Hide progress label
            self.background_label.hide()
            # Enable prompt input field
            self.prompt_input.setEnabled(True)
            self.prompt_input.setFocus()
            # Replace with stop button
            self.request_button_stop()
            # Enable send request button
            self.send_button.setEnabled(True)
        except RuntimeError:
            raise

    def set_status_ready(self):
        try:
            # Set the inference status
            self.is_in_progress = False
            # Set cursor back
            self.setCursor(Qt.CursorShape.ArrowCursor)
            # Enable prompt input field
            self.prompt_input.setEnabled(True)
            self.prompt_input.setFocus()
            # Return request button
            self.request_button_on()
            # Enable save button
            self.save_history_button.setEnabled(True)
            # Hide progress label
            self.background_label.hide()
        except RuntimeError:
            raise

    def request_button_on(self):
        # Sen request button
        self.send_button.clicked.disconnect()
        self.send_button.setText(self.lexemes.get('dialog_button_send_request'))
        self.send_button.setIcon(self.theme_helper.get_icon(
            theme_icon='play-fill.svg',
            color=QColor(self.theme_helper.get_color('ai_assistant_button_icon_color', css_format=True))))
        self.send_button.clicked.connect(self.send_request)

    def request_button_stop(self):
        # Stop request button
        self.send_button.clicked.disconnect()
        self.send_button.setText(self.lexemes.get('dialog_button_stop_request'))
        self.send_button.setIcon(self.theme_helper.get_icon(
            theme_icon='stop-fill.svg',
            color=QColor(self.theme_helper.get_color('ai_assistant_button_icon_color', css_format=True))))
        self.send_button.clicked.connect(self.cancel_request)

    def cancel_request(self):
        self.request_cancelled.emit()

    def save_history_action(self):
        if not hasattr(self.parent, 'action_new_file'):
            self.logger.error("Cannot save history to the file, no method available")
            return

        # Get an instance of the inference module
        module_core = self.init_module()
        # Get a prompt manager
        prompt_manager = module_core.get_prompt_manager()
        # Get history
        history = prompt_manager.get_history()

        # Get current date and time
        current_datetime = datetime.now()
        # Format the date and time in the specified format
        formatted_datetime = current_datetime.strftime("%Y-%m-%d %H:%M:%S")
        # Add header
        history = (self.lexemes.get('dialog_prompt_history_file_header', datetime=formatted_datetime)
                   + f'\n\n{history}')

        # Save history
        self.parent.action_new_file(history)  # noqa

        # Disable save button until next update
        self.save_history_button.setDisabled(True)

    @Slot(int, int)
    def scroll_to_bottom(self, minimum, maximum):
        # Scroll to the bottom of the messages area to show the latest content
        self.messages_scroll.setValue(maximum)

    @asyncClose
    async def closeEvent(self, event):
        self.logger.info('Closing AI Assistant')
        self.dialog_closed.emit()
        self.deleteLater()
        # event.accept()  # Event handled
        super().closeEvent(event)  # Ensure the dialog and base object are closed properly

    def settings_update_handler(self, data) -> None:
        """
        Perform actions upon settings change.
        Data comes in a view of a dictionary, where is the key is the setting name, and the value is the actual value.
        Can be resource greedy.

        Args:
            data dict: say {"settings_key": "..."}

        Return:
            None
        """

        self.logger.debug('Settings update handler is in use "%s"' % data)

        if 'module_ondevice_llm_model_path' in data or 'ai_config_inference_module' in data:
            try:
                if hasattr(self, 'model_name_label'):
                    # Update model name (any updates in settings)
                    self.model_name_label.setText('')
            except RuntimeError:
                # Object can be deleted already
                pass

        if 'app_theme' in data:
            # Update the widget's stylesheet with the selected theme
            self.setStyleSheet(self.theme_helper.get_css('ai_assistant'))



# Korean lexemes toolbar.py
lexemes = {
    "actions_label_new_document": "새 문서",
    "actions_accessible_name_new_document": "새 문서",
    "actions_label_open": "열기",
    "actions_accessible_name_open": "문서 열기",
    "actions_label_save": "저장",
    "actions_accessible_name_save": "문서 저장",
    "actions_label_save_as": "다른 이름으로 저장",
    "actions_accessible_name_save_as": "문서 복사본 저장",
    "actions_label_sync": "동기화",
    "actions_accessible_name_sync": "문서를 외부 저장소와 동기화",
    "actions_label_edit": "편집",
    "actions_accessible_name_edit": "파일 편집",
    "actions_label_source": "소스",
    "actions_accessible_name_source": "소스 코드",
    "actions_label_encrypt": "암호화",
    "actions_accessible_name_encrypt": "파일 암호화",
    "actions_label_decrypt": "복호화",
    "actions_accessible_name_decrypt": "파일 복호화",
    "actions_label_bold": "굵게",
    "actions_accessible_name_bold": "굵은 텍스트 형식",
    "actions_label_italic": "기울임꼴",
    "actions_accessible_name_italic": "기울임꼴 텍스트 형식",
    "actions_label_underline": "밑줄",
    "actions_accessible_name_underline": "밑줄 텍스트 형식",
    "actions_label_strikethrough": "취소선",
    "actions_accessible_name_strikethrough": "취소선 텍스트 형식",
    "actions_label_blockquote": "인용구",
    "actions_accessible_name_blockquote": "인용구 텍스트 형식",
    "actions_label_color_picker": "색상 선택기",
    "actions_accessible_name_color_picker": "텍스트 색상 선택",
    "actions_label_ai_assistant": "AI 보조",
    "actions_accessible_name_ai_assistant": "AI 보조 실행",
    "actions_label_about": "정보",
    "actions_accessible_name_about": "앱 정보",
    "actions_label_exit": "종료",
    "actions_accessible_name_exit": "앱 종료",
    "actions_label_settings": "설정",
    "actions_accessible_name_settings": "앱 설정",

    "search_input_label": "검색",
    "search_input_placeholder_text": "문서에서 텍스트 검색",
    "search_input_accessible_description": "검색 입력 필드",
    "search_buttons_label_clear": "검색 지우기",
    "search_buttons_accessible_name_clear": "검색 입력 필드 지우기",
    "search_buttons_label_prev": "이전 검색",
    "search_buttons_accessible_name_prev": "이전 검색 결과",
    "search_buttons_label_next": "다음 검색",
    "search_buttons_accessible_name_next": "다음 검색 결과",
    "search_case_sensitive_accessible_name": "대소문자 구분",
    "search_case_sensitive_tooltip": "대소문자 구분 검색",
}



# Hindi lexemes ai_assistant.py
lexemes = {
    "dialog_title": "एआई से अनुरोध",
    "dialog_prompt_input_placeholder_text": "अनुरोध",
    "dialog_prompt_input_accessible_description": "एआई-सहायक को संदर्भ प्रदान करने के लिए अनुरोध दर्ज करें",
    "dialog_response_output_placeholder_text": "प्रतिक्रिया डेटा",
    "dialog_response_output_accessible_description": "प्रतिक्रिया डेटा इस फ़ील्ड में प्रदर्शित होगा",
    "dialog_response_output_notice_empty_text": "प्रतिक्रिया प्राप्त करने के लिए पाठ इनपुट प्रदान करें...",

    "dialog_usage_module_label": "मॉड्यूल",
    "dialog_usage_model_label": "मॉडल",
    "dialog_usage_tokens_label": "टोकन",
    "dialog_usage_tokens_prompt": "प्रॉम्प्ट: {tokens}",
    "dialog_usage_tokens_answer": "उत्तर: {tokens}",
    "dialog_usage_tokens_total": "कुल: {tokens}",

    "dialog_button_send_request": "अनुरोध भेजें",
    "dialog_button_stop_request": "रोकें",
    "dialog_button_save_history": "डायलॉग इतिहास सहेजें",

    "dialog_message_copy_tooltip": "पाठ कॉपी करें",
    "dialog_message_copied_tooltip": "कॉपी किया गया",

    "dialog_error_loading_model": "कस्टम मॉडल लोड करने में त्रुटि",
}



# Italian lexemes settings_dialog.py
lexemes = {
    # Impostazioni
    "window_title": "Impostazioni",

    "button_close": "Chiudi",

    "tab_general": "Generale",
    "tab_editor_config": "Editore",
    "tab_viewer_config": "Visualizzatore",
    "tab_ai_config": "Configurazione IA",

    "general_app_config_label": "Configurazione dell'app",
    "general_app_language_label": "Lingua",
    "general_app_language_combo_placeholder_text": "Scegli una lingua",
    "general_app_language_combo_accessible_description": "Lingua dell'interfaccia dell'app",
    "general_app_theme_label": "Tema",
    "general_app_theme_combo_placeholder_text": "Scegli un tema",
    "general_app_theme_combo_accessible_description": "Tema dell'interfaccia dell'app",
    "general_app_main_menu_label": "Menu Principale",
    "general_app_main_menu_checkbox": "Mostra il menu principale",
    "general_app_main_menu_checkbox_accessible_description": "Visualizza il menu a discesa principale dell'app",
    "general_app_font_size_label": "Dimensione del carattere: {size}pt",
    "general_app_font_size_slider_accessible_description": "Regola la dimensione del carattere globale dell'app",

    "general_statusbar_label": "Barra di Stato",
    "general_statusbar_show_global_cursor_position_checkbox": "Mostra la Posizione Globale del Cursore",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "Visualizza la posizione globale del cursore nella barra di stato",

    "editor_config_label": "Configurazione Editor",
    "editor_config_show_line_numbers_checkbox": "Mostra Numeri di Riga",
    "editor_config_show_line_numbers_checkbox_accessible_description": "Visualizza i numeri di riga nell'editor",

    "viewer_config_label": "Configurazione Visualizzatore",
    "viewer_config_process_emojis_checkbox": "Converti Emoji di Testo in Grafica",
    "viewer_config_process_emojis_checkbox_accessible_description":
        "Converti le emoji di testo in rappresentazioni grafiche",
    "viewer_config_highlight_todos_checkbox": "Evidenzia TODO",
    "viewer_config_highlight_todos_checkbox_accessible_description": "Evidenzia le etichette TODO nel testo",
    "viewer_config_open_link_confirmation_checkbox": "Richiedi Conferma per Aprire i Link",
    "viewer_config_open_link_confirmation_checkbox_accessible_description":
        "Richiedi conferma prima di aprire i link",
    "viewer_config_save_resources_checkbox": "Salva automaticamente le immagini esterne sul disco",
    "viewer_config_save_resources_checkbox_accessible_description":
        "Salva automaticamente copie delle immagini esterne sul disco per l'accesso offline.",

    "ai_config_inference_module_label": "Modulo di Inferenza",
    "ai_config_inference_module_names_combo_label": "Modulo di Inferenza Attivo",
    "ai_config_inference_module_names_combo_placeholder_text": "Scegli Modulo",
    "ai_config_inference_module_names_combo_accessible_description":
        "Seleziona tra i moduli di inferenza AI disponibili per operare con l'Assistente AI.\n"
        "Le opzioni includono modelli di linguaggio di grandi dimensioni (LLM) con elaborazione in tempo reale,\n"
        "o funzionalità basate su API.",

    "ai_config_base_label": "Parametri di base",
    "ai_config_multi_turn_dialogue_checkbox": "Dialogo multi-turno con memoria conversazionale",
    "ai_config_multi_turn_dialogue_checkbox_accessible_description":
        "Abilita un dialogo multi-turno che conserva il prompt precedente per la memoria conversazionale.\n"
        "Quando disattivato, solo il nuovo messaggio e il prompt di sistema influenzano la risposta.",
    "ai_config_convert_to_md_checkbox": "Converti il risultato in Markdown",
    "ai_config_convert_to_md_checkbox_accessible_description":
        "Converte il messaggio di output in formato Markdown.",
}



# Japanese lexemes common.py
lexemes = {
    # Common
    "module_ondevice_llm_name": "デバイス上のLLM",

    "module_ondevice_llm_model_exception": "モデルが見つかりませんエラー: {error_msg}",
    "module_ondevice_llm_task_exception": "モデルジェネレータを初期化できません: {error_msg}",
}



# Portuguese lexemes color_picker_dialog.py
lexemes = {
    "color_picker_title": "Selecionar uma cor",

    "color_picker_color_black": "Preto",
    "color_picker_color_silver": "Prata",
    "color_picker_color_gray": "Cinza",
    "color_picker_color_white": "Branco",
    "color_picker_color_maroon": "Bordô",
    "color_picker_color_red": "Vermelho",
    "color_picker_color_purple": "Roxo",
    "color_picker_color_fuchsia": "Fúcsia",
    "color_picker_color_green": "Verde",
    "color_picker_color_lime": "Lima",
    "color_picker_color_olive": "Oliva",
    "color_picker_color_yellow": "Amarelo",
    "color_picker_color_navy": "Azul-marinho",
    "color_picker_color_blue": "Azul",
    "color_picker_color_teal": "Azul-petróleo",
    "color_picker_color_aqua": "Água",
    "color_picker_color_aliceblue": "Azul Alice",
    "color_picker_color_antiquewhite": "Branco Antigo",
    "color_picker_color_aquamarine": "Água-marinha",
    "color_picker_color_azure": "Azul-celeste",
    "color_picker_color_beige": "Bege",
    "color_picker_color_bisque": "Biscoito",
    "color_picker_color_blanchedalmond": "Amêndoa Branqueada",
    "color_picker_color_blueviolet": "Azul-violeta",
    "color_picker_color_brown": "Marrom",
    "color_picker_color_burlywood": "Madeira",
    "color_picker_color_cadetblue": "Azul Cadete",
    "color_picker_color_chartreuse": "Chartreuse",
    "color_picker_color_chocolate": "Chocolate",
    "color_picker_color_coral": "Coral",
    "color_picker_color_cornflowerblue": "Azul Centáurea",
    "color_picker_color_cornsilk": "Seda de Milho",
    "color_picker_color_crimson": "Carmesim",
    "color_picker_color_cyan": "Ciano",
    "color_picker_color_darkblue": "Azul Escuro",
    "color_picker_color_darkcyan": "Ciano Escuro",
    "color_picker_color_darkgoldenrod": "Dourado Escuro",
    "color_picker_color_darkgray": "Cinza Escuro",
    "color_picker_color_darkgreen": "Verde Escuro",
    "color_picker_color_darkkhaki": "Caqui Escuro",
    "color_picker_color_darkmagenta": "Magenta Escuro",
    "color_picker_color_darkolivegreen": "Oliva Escura",
    "color_picker_color_darkorange": "Laranja Escuro",
    "color_picker_color_darkorchid": "Orquídea Escura",
    "color_picker_color_darkred": "Vermelho Escuro",
    "color_picker_color_darksalmon": "Salmão Escuro",
    "color_picker_color_darkseagreen": "Verde-mar Escuro",
    "color_picker_color_darkslateblue": "Azul-ardósia Escuro",
    "color_picker_color_darkslategray": "Cinza-ardósia Escuro",
    "color_picker_color_darkturquoise": "Turquesa Escuro",
    "color_picker_color_darkviolet": "Violeta Escuro",
    "color_picker_color_deeppink": "Rosa Profundo",
    "color_picker_color_deepskyblue": "Azul Céu Profundo",
    "color_picker_color_dimgray": "Cinza Escuro",
    "color_picker_color_dodgerblue": "Azul Dodger",
    "color_picker_color_firebrick": "Tijolo Refratário",
    "color_picker_color_floralwhite": "Branco Floral",
    "color_picker_color_forestgreen": "Verde Floresta",
    "color_picker_color_gainsboro": "Gainsboro",
    "color_picker_color_ghostwhite": "Branco Fantasma",
    "color_picker_color_gold": "Ouro",
    "color_picker_color_goldenrod": "Dourado",
    "color_picker_color_greenyellow": "Verde Amarelado",
    "color_picker_color_honeydew": "Mel",
    "color_picker_color_hotpink": "Rosa Choque",
    "color_picker_color_indianred": "Vermelho Índio",
    "color_picker_color_indigo": "Índigo",
    "color_picker_color_ivory": "Marfim",
    "color_picker_color_khaki": "Caqui",
    "color_picker_color_lavender": "Lavanda",
    "color_picker_color_lavenderblush": "Lavanda Rosada",
    "color_picker_color_lawngreen": "Verde Grama",
    "color_picker_color_lemonchiffon": "Chiffon de Limão",
    "color_picker_color_lightblue": "Azul Claro",
    "color_picker_color_lightcoral": "Coral Claro",
    "color_picker_color_lightcyan": "Ciano Claro",
    "color_picker_color_lightgoldenrodyellow": "Amarelo Ouro Claro",
    "color_picker_color_lightgray": "Cinza Claro",
    "color_picker_color_lightgreen": "Verde Claro",
    "color_picker_color_lightpink": "Rosa Claro",
    "color_picker_color_lightsalmon": "Salmão Claro",
    "color_picker_color_lightseagreen": "Verde-mar Claro",
    "color_picker_color_lightskyblue": "Azul Céu Claro",
    "color_picker_color_lightslategray": "Cinza-ardósia Claro",
    "color_picker_color_lightsteelblue": "Azul Aço Claro",
    "color_picker_color_lightyellow": "Amarelo Claro",
    "color_picker_color_limegreen": "Verde Lima",
    "color_picker_color_linen": "Linho",
    "color_picker_color_magenta": "Magenta",
    "color_picker_color_mediumaquamarine": "Água-marinha Médio",
    "color_picker_color_mediumblue": "Azul Médio",
    "color_picker_color_mediumorchid": "Orquídea Média",
    "color_picker_color_mediumpurple": "Púrpura Médio",
    "color_picker_color_mediumseagreen": "Verde-mar Médio",
    "color_picker_color_mediumslateblue": "Azul-ardósia Médio",
    "color_picker_color_mediumspringgreen": "Verde Primavera Médio",
    "color_picker_color_mediumturquoise": "Turquesa Médio",
    "color_picker_color_mediumvioletred": "Vermelho-violeta Médio",
    "color_picker_color_midnightblue": "Azul Meia-noite",
    "color_picker_color_mintcream": "Creme de Menta",
    "color_picker_color_mistyrose": "Rosa Névoa",
    "color_picker_color_moccasin": "Mocassim",
    "color_picker_color_navajowhite": "Branco Navajo",
    "color_picker_color_oldlace": "Renda Antiga",
    "color_picker_color_olivedrab": "Oliva Parda",
    "color_picker_color_orange": "Laranja",
    "color_picker_color_orangered": "Vermelho-laranja",
    "color_picker_color_orchid": "Orquídea",
    "color_picker_color_palegoldenrod": "Dourado Pálido",
    "color_picker_color_palegreen": "Verde Pálido",
    "color_picker_color_paleturquoise": "Turquesa Pálido",
    "color_picker_color_palevioletred": "Vermelho-violeta Pálido",
    "color_picker_color_papayawhip": "Creme de Papaia",
    "color_picker_color_peachpuff": "Pêssego",
    "color_picker_color_peru": "Peru",
    "color_picker_color_pink": "Rosa",
    "color_picker_color_plum": "Ameixa",
    "color_picker_color_powderblue": "Azul Pó",
    "color_picker_color_rebeccapurple": "Roxo Rebecca",
    "color_picker_color_rosybrown": "Marrom Rosado",
    "color_picker_color_royalblue": "Azul Real",
    "color_picker_color_saddlebrown": "Marrom Sela",
    "color_picker_color_salmon": "Salmão",
    "color_picker_color_sandybrown": "Marrom Areia",
    "color_picker_color_seagreen": "Verde-mar",
    "color_picker_color_seashell": "Concha",
    "color_picker_color_sienna": "Siena",
    "color_picker_color_skyblue": "Azul Céu",
    "color_picker_color_slateblue": "Azul Ardósia",
    "color_picker_color_slategray": "Cinza Ardósia",
    "color_picker_color_snow": "Neve",
    "color_picker_color_springgreen": "Verde Primavera",
    "color_picker_color_steelblue": "Azul Aço",
    "color_picker_color_tan": "Bronzeado",
    "color_picker_color_thistle": "Cardo",
    "color_picker_color_tomato": "Tomate",
    "color_picker_color_turquoise": "Turquesa",
    "color_picker_color_violet": "Violeta",
    "color_picker_color_wheat": "Trigo",
    "color_picker_color_whitesmoke": "Fumaça Branca",
    "color_picker_color_yellowgreen": "Verde-amarelo",
}
